# TF general functions
from PrognosAIs.Model.Architectures.Architecture import ClassificationNetworkArchitecture
from tensorflow.keras.backend import int_shape

# TF 3D functions
# TF 2D functions
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Concatenate
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Conv3D
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import GlobalAveragePooling2D
from tensorflow.keras.layers import GlobalAveragePooling3D
from tensorflow.keras.layers import MaxPooling2D
from tensorflow.keras.layers import MaxPooling3D
from tensorflow.keras.layers import ReLU
from tensorflow.keras.models import Model


# Other


class DenseNet(ClassificationNetworkArchitecture):
    def init_dimensionality(self, N_dimension):
        if N_dimension == 2:
            self.dims = 2
            self.conv_func = Conv2D
            self.pool_func = MaxPooling2D
        elif N_dimension == 3:
            self.dims = 3
            self.conv_func = Conv3D
            self.pool_func = MaxPooling3D

    def get_dense_stem(self, layer, N_filters):
        stride_size = self.get_corrected_stride_size(layer, [2] * self.dims, [7] * self.dims)

        conv_1 = self.conv_func(
            filters=N_filters,
            kernel_size=[7] * self.dims,
            strides=stride_size,
            padding="same",
            data_format="channels_last",
            activation="relu",
        )(layer)

        stride_size = self.get_corrected_stride_size(conv_1, [2] * self.dims, [3] * self.dims)
        pooling_1 = self.pool_func(
            pool_size=[3] * self.dims,
            strides=stride_size,
            padding="same",
            data_format="channels_last",
        )(conv_1)

        return pooling_1

    def get_dense_block(self, layer, N_filters, N_conv_layers):
        for i_conv_layer in range(N_conv_layers):
            head = BatchNormalization(axis=-1)(layer)
            head = ReLU()(head)
            head = self.conv_func(
                filters=N_filters * 4,
                kernel_size=[1] * self.dims,
                strides=[1] * self.dims,
                padding="same",
                data_format="channels_last",
                activation="linear",
            )(head)

            head = BatchNormalization(axis=-1)(head)
            head = ReLU()(head)
            head = self.conv_func(
                filters=N_filters,
                kernel_size=[3] * self.dims,
                strides=[1] * self.dims,
                padding="same",
                data_format="channels_last",
                activation="linear",
            )(head)

            head = Concatenate(axis=-1)([layer, head])
            layer = head

        return head

    def get_transition_block(self, layer, N_filters, theta):
        conv_1 = self.conv_func(
            filters=int(N_filters * theta),
            kernel_size=[1] * self.dims,
            strides=[1] * self.dims,
            padding="same",
            data_format="channels_last",
            activation="linear",
        )(layer)

        stride_size = self.get_corrected_stride_size(conv_1, [2] * self.dims, [2] * self.dims)

        pooling_1 = self.pool_func(
            pool_size=[2] * self.dims,
            strides=stride_size,
            padding="valid",
            data_format="channels_last",
        )(conv_1)

        return pooling_1


class DenseNet_121_2D(DenseNet):
    dims = 2
    GROWTH_RATE = 32
    INITIAL_FILTERS = 2 * GROWTH_RATE
    # Settings theta to 1 will create a DenseNet-B
    THETA = 0.5

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [21, 21])

        dense_stem = self.get_dense_stem(self.inputs, self.INITIAL_FILTERS)
        dense_block_1 = self.get_dense_block(dense_stem, self.GROWTH_RATE, 6)
        N_filters = int_shape(dense_block_1)[-1]

        dense_block_1 = self.make_dropout_layer(dense_block_1)
        trans_block_1 = self.get_transition_block(dense_block_1, N_filters, self.THETA)
        dense_block_2 = self.get_dense_block(trans_block_1, self.GROWTH_RATE, 12)
        N_filters = int_shape(dense_block_2)[-1]

        dense_block_2 = self.make_dropout_layer(dense_block_2)
        trans_block_2 = self.get_transition_block(dense_block_2, N_filters, self.THETA)
        dense_block_3 = self.get_dense_block(trans_block_2, self.GROWTH_RATE, 24)
        N_filters = int_shape(dense_block_3)[-1]

        dense_block_3 = self.make_dropout_layer(dense_block_3)
        trans_block_3 = self.get_transition_block(dense_block_3, N_filters, self.THETA)
        dense_block_4 = self.get_dense_block(trans_block_3, self.GROWTH_RATE, 16)

        gap_1 = GlobalAveragePooling2D(data_format="channels_last")(dense_block_4)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class DenseNet_121_3D(DenseNet):
    dims = 3
    GROWTH_RATE = 32
    INITIAL_FILTERS = 2 * GROWTH_RATE
    # Settings theta to 1 will create a DenseNet-B
    THETA = 0.5

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [21, 21, 21])

        dense_stem = self.get_dense_stem(self.inputs, self.INITIAL_FILTERS)
        dense_block_1 = self.get_dense_block(dense_stem, self.GROWTH_RATE, 6)
        N_filters = int_shape(dense_block_1)[-1]

        dense_block_1 = self.make_dropout_layer(dense_block_1)
        trans_block_1 = self.get_transition_block(dense_block_1, N_filters, self.THETA)
        dense_block_2 = self.get_dense_block(trans_block_1, self.GROWTH_RATE, 12)
        N_filters = int_shape(dense_block_2)[-1]

        dense_block_2 = self.make_dropout_layer(dense_block_2)
        trans_block_2 = self.get_transition_block(dense_block_2, N_filters, self.THETA)
        dense_block_3 = self.get_dense_block(trans_block_2, self.GROWTH_RATE, 24)
        N_filters = int_shape(dense_block_3)[-1]

        dense_block_3 = self.make_dropout_layer(dense_block_3)
        trans_block_3 = self.get_transition_block(dense_block_3, N_filters, self.THETA)
        dense_block_4 = self.get_dense_block(trans_block_3, self.GROWTH_RATE, 16)

        gap_1 = GlobalAveragePooling3D(data_format="channels_last")(dense_block_4)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class DenseNet_169_2D(DenseNet):
    dims = 2
    GROWTH_RATE = 32
    INITIAL_FILTERS = 2 * GROWTH_RATE
    # Settings theta to 1 will create a DenseNet-B
    THETA = 0.5

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [21, 21])

        dense_stem = self.get_dense_stem(self.inputs, self.INITIAL_FILTERS)
        dense_block_1 = self.get_dense_block(dense_stem, self.GROWTH_RATE, 6)
        N_filters = int_shape(dense_block_1)[-1]

        dense_block_1 = self.make_dropout_layer(dense_block_1)
        trans_block_1 = self.get_transition_block(dense_block_1, N_filters, self.THETA)
        dense_block_2 = self.get_dense_block(trans_block_1, self.GROWTH_RATE, 12)
        N_filters = int_shape(dense_block_2)[-1]

        dense_block_2 = self.make_dropout_layer(dense_block_2)
        trans_block_2 = self.get_transition_block(dense_block_2, N_filters, self.THETA)
        dense_block_3 = self.get_dense_block(trans_block_2, self.GROWTH_RATE, 32)
        N_filters = int_shape(dense_block_3)[-1]

        dense_block_3 = self.make_dropout_layer(dense_block_3)
        trans_block_3 = self.get_transition_block(dense_block_3, N_filters, self.THETA)
        dense_block_4 = self.get_dense_block(trans_block_3, self.GROWTH_RATE, 32)

        gap_1 = GlobalAveragePooling2D(data_format="channels_last")(dense_block_4)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class DenseNet_169_3D(DenseNet):
    dims = 3
    GROWTH_RATE = 32
    INITIAL_FILTERS = 2 * GROWTH_RATE
    # Settings theta to 1 will create a DenseNet-B
    THETA = 0.5

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [21, 21, 21])

        dense_stem = self.get_dense_stem(self.inputs, self.INITIAL_FILTERS)
        dense_block_1 = self.get_dense_block(dense_stem, self.GROWTH_RATE, 6)
        N_filters = int_shape(dense_block_1)[-1]

        dense_block_1 = self.make_dropout_layer(dense_block_1)
        trans_block_1 = self.get_transition_block(dense_block_1, N_filters, self.THETA)
        dense_block_2 = self.get_dense_block(trans_block_1, self.GROWTH_RATE, 12)
        N_filters = int_shape(dense_block_2)[-1]

        dense_block_2 = self.make_dropout_layer(dense_block_2)
        trans_block_2 = self.get_transition_block(dense_block_2, N_filters, self.THETA)
        dense_block_3 = self.get_dense_block(trans_block_2, self.GROWTH_RATE, 32)
        N_filters = int_shape(dense_block_3)[-1]

        dense_block_2 = self.make_dropout_layer(dense_block_2)
        trans_block_3 = self.get_transition_block(dense_block_3, N_filters, self.THETA)
        dense_block_4 = self.get_dense_block(trans_block_3, self.GROWTH_RATE, 32)

        gap_1 = GlobalAveragePooling3D(data_format="channels_last")(dense_block_4)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class DenseNet_201_2D(DenseNet):
    dims = 2
    GROWTH_RATE = 32
    INITIAL_FILTERS = 2 * GROWTH_RATE
    # Settings theta to 1 will create a DenseNet-B
    THETA = 0.5

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [21, 21])

        dense_stem = self.get_dense_stem(self.inputs, self.INITIAL_FILTERS)
        dense_block_1 = self.get_dense_block(dense_stem, self.GROWTH_RATE, 6)
        N_filters = int_shape(dense_block_1)[-1]

        dense_block_1 = self.make_dropout_layer(dense_block_1)
        trans_block_1 = self.get_transition_block(dense_block_1, N_filters, self.THETA)
        dense_block_2 = self.get_dense_block(trans_block_1, self.GROWTH_RATE, 12)
        N_filters = int_shape(dense_block_2)[-1]

        dense_block_2 = self.make_dropout_layer(dense_block_2)
        trans_block_2 = self.get_transition_block(dense_block_2, N_filters, self.THETA)
        dense_block_3 = self.get_dense_block(trans_block_2, self.GROWTH_RATE, 48)
        N_filters = int_shape(dense_block_3)[-1]

        dense_block_3 = self.make_dropout_layer(dense_block_3)
        trans_block_3 = self.get_transition_block(dense_block_3, N_filters, self.THETA)
        dense_block_4 = self.get_dense_block(trans_block_3, self.GROWTH_RATE, 48)

        gap_1 = GlobalAveragePooling2D(data_format="channels_last")(dense_block_4)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class DenseNet_201_3D(DenseNet):
    dims = 3
    GROWTH_RATE = 32
    INITIAL_FILTERS = 2 * GROWTH_RATE
    # Settings theta to 1 will create a DenseNet-B
    THETA = 0.5

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [21, 21, 21])

        dense_stem = self.get_dense_stem(self.inputs, self.INITIAL_FILTERS)
        dense_block_1 = self.get_dense_block(dense_stem, self.GROWTH_RATE, 6)
        N_filters = int_shape(dense_block_1)[-1]

        dense_block_1 = self.make_dropout_layer(dense_block_1)
        trans_block_1 = self.get_transition_block(dense_block_1, N_filters, self.THETA)
        dense_block_2 = self.get_dense_block(trans_block_1, self.GROWTH_RATE, 12)
        N_filters = int_shape(dense_block_2)[-1]

        dense_block_2 = self.make_dropout_layer(dense_block_2)
        trans_block_2 = self.get_transition_block(dense_block_2, N_filters, self.THETA)
        dense_block_3 = self.get_dense_block(trans_block_2, self.GROWTH_RATE, 48)
        N_filters = int_shape(dense_block_3)[-1]

        dense_block_3 = self.make_dropout_layer(dense_block_3)
        trans_block_3 = self.get_transition_block(dense_block_3, N_filters, self.THETA)
        dense_block_4 = self.get_dense_block(trans_block_3, self.GROWTH_RATE, 48)

        gap_1 = GlobalAveragePooling3D(data_format="channels_last")(dense_block_4)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class DenseNet_264_2D(DenseNet):
    dims = 2
    GROWTH_RATE = 32
    INITIAL_FILTERS = 2 * GROWTH_RATE
    # Settings theta to 1 will create a DenseNet-B
    THETA = 0.5

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [21, 21])

        dense_stem = self.get_dense_stem(self.inputs, self.INITIAL_FILTERS)
        dense_block_1 = self.get_dense_block(dense_stem, self.GROWTH_RATE, 6)
        N_filters = int_shape(dense_block_1)[-1]

        dense_block_1 = self.make_dropout_layer(dense_block_1)
        trans_block_1 = self.get_transition_block(dense_block_1, N_filters, self.THETA)
        dense_block_2 = self.get_dense_block(trans_block_1, self.GROWTH_RATE, 12)
        N_filters = int_shape(dense_block_2)[-1]

        dense_block_2 = self.make_dropout_layer(dense_block_2)
        trans_block_2 = self.get_transition_block(dense_block_2, N_filters, self.THETA)
        dense_block_3 = self.get_dense_block(trans_block_2, self.GROWTH_RATE, 64)
        N_filters = int_shape(dense_block_3)[-1]

        dense_block_3 = self.make_dropout_layer(dense_block_3)
        trans_block_3 = self.get_transition_block(dense_block_3, N_filters, self.THETA)
        dense_block_4 = self.get_dense_block(trans_block_3, self.GROWTH_RATE, 48)

        gap_1 = GlobalAveragePooling2D(data_format="channels_last")(dense_block_4)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model


class DenseNet_264_3D(DenseNet):
    dims = 3
    GROWTH_RATE = 32
    INITIAL_FILTERS = 2 * GROWTH_RATE
    # Settings theta to 1 will create a DenseNet-B
    THETA = 0.5

    def create_model(self):
        self.init_dimensionality(self.dims)
        self.check_minimum_input_size(self.inputs, [21, 21, 21])

        dense_stem = self.get_dense_stem(self.inputs, self.INITIAL_FILTERS)
        dense_block_1 = self.get_dense_block(dense_stem, self.GROWTH_RATE, 6)
        N_filters = int_shape(dense_block_1)[-1]

        dense_block_1 = self.make_dropout_layer(dense_block_1)
        trans_block_1 = self.get_transition_block(dense_block_1, N_filters, self.THETA)
        dense_block_2 = self.get_dense_block(trans_block_1, self.GROWTH_RATE, 12)
        N_filters = int_shape(dense_block_2)[-1]

        dense_block_2 = self.make_dropout_layer(dense_block_2)
        trans_block_2 = self.get_transition_block(dense_block_2, N_filters, self.THETA)
        dense_block_3 = self.get_dense_block(trans_block_2, self.GROWTH_RATE, 64)
        N_filters = int_shape(dense_block_3)[-1]

        dense_block_3 = self.make_dropout_layer(dense_block_3)
        trans_block_3 = self.get_transition_block(dense_block_3, N_filters, self.THETA)
        dense_block_4 = self.get_dense_block(trans_block_3, self.GROWTH_RATE, 48)

        gap_1 = GlobalAveragePooling3D(data_format="channels_last")(dense_block_4)

        predictions = self.outputs(gap_1)

        model = Model(inputs=self.inputs, outputs=predictions)

        return model



import argparse
import copy
import itertools
import json
import os
import shutil
import sys

from multiprocessing import Pool
from typing import Tuple
from typing import Union

import h5py
import numpy as np
import PrognosAIs.IO.ConfigLoader
import PrognosAIs.IO.Configs
import PrognosAIs.IO.LabelParser
import PrognosAIs.IO.utils as IO_utils
import SimpleITK as sitk
import sklearn.model_selection

from PrognosAIs.Preprocessing import Samples
from PrognosAIs.Preprocessing.Samples import ImageSample


class SingleSamplePreprocessor:
    def __init__(
        self, sample: ImageSample, config: dict, output_directory: str = None,
    ):
        # We make sure we dont change to the original objects
        self.sample = sample.copy()
        self.config = copy.deepcopy(config)
        self.output_directory = output_directory
        self.save_names = None

        self.general_config = PrognosAIs.IO.Configs.general_config(self.config)

        self._HDF5_EXTENSION = ".hdf5"
        self._PATCH_SEPARATOR = "_patch_"

        self._init_configs()

    def _init_configs(self):
        for i_possible_step in self.general_config.pipeline:
            config_name = i_possible_step + "_config"
            config_class = getattr(PrognosAIs.IO.Configs, config_name)

            if i_possible_step in self.config:
                setattr(self, config_name, config_class(self.config[i_possible_step]))
            else:
                setattr(self, config_name, config_class(None))

    def build_pipeline(self) -> list:
        # TODO add potential extra inputs
        # TODO add reorienting to standard space
        # TODO registration of different channels in sample
        # TODO RGB rejecting/to gray

        pipeline_image = []
        pipeline_patches = []

        for i_possible_step in self.general_config.pipeline:
            i_step_config = getattr(self, i_possible_step + "_config")
            if i_step_config.perform_step and i_step_config.perform_step_on_image:
                pipeline_image.append(getattr(self, i_possible_step))
            elif i_step_config.perform_step and i_step_config.perform_step_on_patch:
                pipeline_patches.append(getattr(self, i_possible_step))

        pipeline = pipeline_image + pipeline_patches

        return pipeline

    def apply_pipeline(self, pipeline=None):
        if pipeline is None:
            pipeline = self.build_pipeline()

        for i_step in pipeline:
            success = i_step()
            if success is False:
                break

    # ===============================================================
    # Image dimension extraction
    # ===============================================================
    def multi_dimension_extracting(self):
        """
        Extract invidiual images from a multi-dimensional sequence.

        Raises:
            NotImplementedError: If an extraction type is requested that is not supported.
        """
        if (
            self.multi_dimension_extracting_config.max_number_of_dimensions
            < self.sample.number_of_dimensions
        ):
            max_dims = self.multi_dimension_extracting_config.max_number_of_dimensions
            if self.multi_dimension_extracting_config.extraction_type == "first":
                extraction_fuction = self._get_first_image_from_sequence
            elif self.multi_dimension_extracting_config.extraction_type == "all":
                extraction_fuction = self._get_all_images_from_sequence
            else:
                raise NotImplementedError(
                    "Unknown extraction type {}".format(
                        self.multi_dimension_extracting_config.extraction_type
                    )
                )

            self.sample.channels = (extraction_fuction, [max_dims])
            if self.multi_dimension_extracting_config.extract_masks:
                self.sample.masks = (extraction_fuction, [max_dims])

    @staticmethod
    def _get_first_image_from_sequence(image: sitk.Image, max_dims: int) -> sitk.Image:
        """
        Extract the first image from a sequence of images

        Args:
            image (sitk.Image): Multi-dimensional image containg the sequence.
            max_dims (int): The maximum number of dimension the output can be.

        Returns:
            sitk.Image: The first image extracted from the sequence
        """

        image_size = list(image.GetSize())
        image_dims = len(image_size)
        to_cut_dim = image_dims - max_dims

        image_size[max_dims:] = [0] * to_cut_dim
        image = sitk.Extract(image, size=image_size, index=[0] * image_dims)

        return image

    @staticmethod
    def _get_all_images_from_sequence(image: sitk.Image, max_dims: int) -> list:
        """
        Get all of the images from a sequence of images.

        Args:
            image (sitk.Image): Multi-dimensional image containg the sequence.
            max_dims (int): The number of dimension of each individual image.
             This should be equal to the dimensionality of the input image - 1.
             Otherwise, we do not know how to extract the appropiate images

        Raises:
            ValueError: If the maximum number of dimensions does not fit with the sequences.

        Returns:
            list: All images extracted from the sequence.
        """

        image_size = list(image.GetSize())
        image_dims = len(image_size)

        if max_dims + 1 != image_dims:
            err_msg = """When extracting all dimensions of an image, the image can only have one
            more dimension than the max dimension.
            Image had {} dimensions, but max dimensions was set to {}."""
            raise ValueError(err_msg.format(image_dims, max_dims))

        image_extractor = sitk.ExtractImageFilter()
        N_patches = image_size[-1]
        patches = []
        image_size[-1] = 0
        image_extractor.SetSize(image_size)
        for i_patch in range(N_patches):
            image_extractor.SetIndex([0] * max_dims + [i_patch])
            patches.append(image_extractor.Execute(image))

        return patches

    # ===============================================================
    # Masking functions
    # ===============================================================

    def masking(self):
        if self.masking_config.mask_background:
            self.mask_background(
                self.masking_config.mask,
                self.masking_config.background_value,
                self.masking_config.process_masks,
            )
        if self.masking_config.crop_to_mask:
            self.crop_to_mask(self.masking_config.mask, self.masking_config.process_masks)

    def mask_background(
        self, ROI_mask: sitk.Image, background_value: float = 0.0, process_masks: bool = True,
    ):
        mask_image_filter = sitk.MaskImageFilter()

        mask_image_filter.SetMaskingValue(0)
        if background_value == "min":
            self.sample.channels = (self.mask_background_to_min, [ROI_mask])
        else:
            mask_image_filter.SetOutsideValue(background_value)
            self.sample.channels = (mask_image_filter.Execute, [ROI_mask])
        if process_masks:
            # background_dtype = ImageSample.get_appropiate_dtype_from_scalar(background_value)
            # if background_dtype != self.sample.get_example_mask().GetPixelID():
            #     common_type = ImageSample.promote_simpleitk_types(
            #         background_dtype, self.sample.get_example_mask().GetPixelID()
            #     )
            #     self.sample.masks = (sitk.Cast, [common_type])
            # TODO fix here as well, we set 0 automatically, but perhaps need to fix this
            mask_image_filter.SetOutsideValue(0.0)
            self.sample.masks = (mask_image_filter.Execute, [ROI_mask])

    @staticmethod
    def mask_background_to_min(image, mask):
        mask_label_filter = sitk.LabelIntensityStatisticsImageFilter()
        mask_label_filter.Execute(mask, image)

        img_min = mask_label_filter.GetMinimum(1)

        image = sitk.Mask(image, mask, img_min)
        return image

    def crop_to_mask(self, ROI_mask: sitk.Image, process_masks: bool = True):
        statics_image_filter = sitk.LabelShapeStatisticsImageFilter()
        statics_image_filter.Execute(ROI_mask)

        mask_bounding_box = statics_image_filter.GetBoundingBox(1)
        N_dimensions = int(len(mask_bounding_box) / 2)
        bounding_box_index = mask_bounding_box[0:N_dimensions]
        bounding_box_size = mask_bounding_box[N_dimensions:]

        self.sample.channels = (
            sitk.RegionOfInterest,
            {"index": bounding_box_index, "size": bounding_box_size},
        )
        if process_masks:
            self.sample.masks = (
                sitk.RegionOfInterest,
                {"index": bounding_box_index, "size": bounding_box_size},
            )

    # ===============================================================
    # Resampling functions
    # ===============================================================
    def resampling(self):
        channel_resampler = sitk.ResampleImageFilter()
        channel_resampler.SetInterpolator(sitk.sitkBSpline)

        mask_resampler = sitk.ResampleImageFilter()
        mask_resampler.SetInterpolator(sitk.sitkNearestNeighbor)

        self.sample.channels = (
            self._resample,
            [self.resampling_config.resample_size, channel_resampler],
        )
        self.sample.masks = (
            self._resample,
            [self.resampling_config.resample_size, mask_resampler],
        )

    @staticmethod
    def _resample(image, resample_size, resampler):
        original_size = np.asarray(image.GetSize())
        original_spacing = np.asarray(image.GetSpacing())
        resample_size = np.asarray(resample_size)

        new_spacing = original_size * original_spacing / resample_size

        resampler.SetOutputSpacing(new_spacing.tolist())
        resampler.SetSize(resample_size.tolist())

        resampler.SetOutputDirection(image.GetDirection())
        resampler.SetOutputOrigin(image.GetOrigin())
        resampler.SetOutputPixelType(image.GetPixelID())
        resampler.SetTransform(sitk.Transform())

        image = resampler.Execute(image)
        return image

    # ===============================================================
    # Normalizing functions
    # ===============================================================
    def normalizing(self):
        if (
            self.normalizing_config.normalization_method == "range"
            and self.normalizing_config.mask is None
        ):
            self.sample.channels = (
                self._rescale_image_intensity_range,
                [
                    self.normalizing_config.normalization_range,
                    self.normalizing_config.output_range,
                ],
            )
        elif (
            self.normalizing_config.normalization_method == "range"
            and self.normalizing_config.mask is not None
        ):
            self.sample.channels = (
                self._rescale_image_intensity_range_with_mask,
                [
                    self.normalizing_config.mask,
                    self.normalizing_config.normalization_range,
                    self.normalizing_config.output_range,
                ],
            )
        elif (
            self.normalizing_config.normalization_method == "zscore"
            and self.normalizing_config.mask is None
        ):
            self.sample.channels = self._zscore_image_intensity
        elif (
            self.normalizing_config.normalization_method == "zscore"
            and self.normalizing_config.mask is not None
        ):
            self.sample.channels = (
                self._zscore_image_intensity_with_mask,
                [self.normalizing_config.mask],
            )

        if self.normalizing_config.mask_normalization == "collapse":
            self.sample.masks = self._collapse_mask
        elif self.normalizing_config.mask_normalization == "consecutively":
            self.sample.masks = self._make_consecutive_mask

        if self.normalizing_config.mask_smoothing:
            self.sample.masks = self._smooth_mask

    @staticmethod
    def _rescale_image_intensity_range(
        image: sitk.Image, percentile_range: list, output_range: list = None
    ) -> sitk.Image:
        image_array = sitk.GetArrayViewFromImage(image)

        low_intensity = np.percentile(image_array, percentile_range[0])
        high_intensity = np.percentile(image_array, percentile_range[1])
        if np.isclose(low_intensity, high_intensity):
            raise ValueError(
                """Percentiles are too close, or image intensity is
         too imbalanced, cannot normalize"""
            )

        image = sitk.IntensityWindowing(
            image, low_intensity, high_intensity, low_intensity, high_intensity
        )
        if output_range is not None:
            image = sitk.RescaleIntensity(image, output_range[0], output_range[1])
        return image

    @staticmethod
    def _rescale_image_intensity_range_with_mask(
        image: sitk.Image, mask: sitk.Image, percentile_range: list, output_range: list = None,
    ) -> sitk.Image:

        image_array = sitk.GetArrayViewFromImage(image)
        mask_array = sitk.GetArrayViewFromImage(mask)

        masked_values = image_array[mask_array > 0].flatten()

        low_intensity = np.percentile(masked_values, percentile_range[0])
        high_intensity = np.percentile(masked_values, percentile_range[1])

        if np.isclose(low_intensity, high_intensity):
            raise ValueError(
                """Percentiles are too close, or image intensity is
         too imbalanced, cannot normalize"""
            )

        image = sitk.IntensityWindowing(
            image, low_intensity, high_intensity, low_intensity, high_intensity
        )

        if output_range is not None:
            image = sitk.RescaleIntensity(image, output_range[0], output_range[1])
        return image

    @staticmethod
    def _zscore_image_intensity(image: sitk.Image) -> sitk.Image:
        return sitk.Normalize(image)

    @staticmethod
    def _zscore_image_intensity_with_mask(image: sitk.Image, mask: sitk.Image) -> sitk.Image:
        mask_label_filter = sitk.LabelIntensityStatisticsImageFilter()
        mask_label_filter.Execute(mask, image)

        img_mean = mask_label_filter.GetMean(1)
        img_std = mask_label_filter.GetStandardDeviation(1)

        image = sitk.ShiftScale(image, -1.0 * img_mean, 1.0 / img_std)
        return image

    @staticmethod
    def _make_mask_positive(mask: sitk.Image) -> sitk.Image:
        original_mask = sitk.Image(mask)
        # Make the mask positive, but keep 0 as 0
        minmax_filter = sitk.MinimumMaximumImageFilter()
        minmax_filter.Execute(mask)
        # We make sure that everything is positive
        mask_min = minmax_filter.GetMinimum()
        if mask_min < 0:
            # First we make sure that we dont lose any data in the new data type
            new_max = minmax_filter.GetMaximum() - mask_min
            new_data_type = ImageSample.get_appropiate_dtype_from_scalar(new_max)
            mask = sitk.Subtract(mask, mask_min - 1)
            # Multiply here so what was originally 0 is still 0
            mask = sitk.Multiply(
                mask, sitk.Cast(sitk.NotEqual(original_mask, 0), mask.GetPixelID())
            )
            mask = sitk.Cast(mask, new_data_type)

        return mask

    @staticmethod
    def _collapse_mask(mask: sitk.Image) -> sitk.Image:
        mask = SingleSamplePreprocessor._make_mask_positive(mask)
        mask = sitk.LabelImageToLabelMap(mask)
        mask = sitk.AggregateLabelMap(mask)
        mask = sitk.RelabelLabelMap(mask)
        mask = sitk.LabelMapToLabel(mask)
        # Since we collapsed, it for sure fits into a uint8
        # As there are only two values, so we cast here to save memory
        mask = sitk.Cast(mask, sitk.sitkUInt8)

        return mask

    @staticmethod
    def _smooth_mask(mask: sitk.Image) -> sitk.Image:
        mask = sitk.BinaryMedian(mask, [3, 3, 3])
        return mask

    @staticmethod
    def _make_consecutive_mask(mask: sitk.Image) -> sitk.Image:
        mask = SingleSamplePreprocessor._make_mask_positive(mask)
        mask = sitk.LabelImageToLabelMap(mask)
        mask = sitk.RelabelLabelMap(mask)
        mask = sitk.LabelMapToLabel(mask)

        return mask

    # ===============================================================
    # Padding
    # ===============================================================
    @staticmethod
    def _pad_image_to_size(
        image: sitk.Image, output_size: list, pad_constant: float = 0.0
    ) -> Tuple[sitk.Image, np.ndarray, np.ndarray]:
        image_size = np.asarray(image.GetSize())
        output_size = np.asarray(output_size)
        if any(image_size < output_size):
            required_padding = np.maximum(output_size - image_size, 0)
            left_padding = np.floor(required_padding / 2.0) + np.mod(required_padding, 2.0)
            right_padding = np.floor(required_padding / 2.0)
            image = SingleSamplePreprocessor._pad_image_from_parameters(
                image, left_padding, right_padding, pad_constant
            )
        else:
            left_padding = np.zeros(len(image_size))
            right_padding = np.zeros_like(left_padding)
        return image, left_padding, right_padding

    @staticmethod
    def _pad_image_from_parameters(
        image: sitk.Image, left_padding: np.ndarray, right_padding: np.ndarray, pad_constant: float,
    ) -> sitk.Image:

        left_padding = left_padding.astype(np.int).tolist()
        right_padding = right_padding.astype(np.int).tolist()
        image = sitk.ConstantPad(image, left_padding, right_padding, pad_constant)
        return image

    # ===============================================================
    # Patching
    # ===============================================================
    def patching(self) -> None:
        patch_parameters = self._get_patch_parameters()

        self.sample.channels = (
            self._make_patches,
            [patch_parameters, self.patching_config.pad_constant, self.patching_config.patch_size],
        )
        # TODO this pad constant is 0, because that makes sense for
        # Masks but perhaps lets let users set it themselves
        self.sample.masks = (
            self._make_patches,
            [patch_parameters, 0, self.patching_config.patch_size],
        )

    def _get_patch_parameters(self) -> dict:
        patch_parameters = {}
        patch_parameters["left_padding"] = np.zeros(self.sample.number_of_dimensions)
        patch_parameters["right_padding"] = np.zeros(self.sample.number_of_dimensions)
        patch_parameters["patch_indices"] = None

        # Make sure that all samples are the same size, otherwise the patches
        # wont make sense
        self.sample.assert_all_channels_same_size()
        self.sample.assert_all_masks_same_size()

        example_sample = self.sample.get_example_channel()

        if self.patching_config.pad_if_needed:
            example_sample, left_padding, right_padding = self._pad_image_to_size(
                example_sample, self.patching_config.patch_size, self.patching_config.pad_constant,
            )
            patch_parameters["left_padding"] += left_padding
            patch_parameters["right_padding"] += right_padding

        elif any(example_sample.GetSize() < self.patching_config.patch_size):
            raise ValueError(
                """The sample is smaller than the patch and padding not requested,
                 cannot make patches."""
            )


        if self.patching_config.extraction_type not in ["random", "fitting", "overlap"]:
            raise NotImplementedError(
                "The specified extraction type {} is not specified!".format(
                    self.patching_config.extraction_type
                )
            )

        if self.patching_config.extraction_type == "random":
            patch_parameters["patch_indices"] = self._get_random_patching_parameters(
                self.patching_config.patch_size,
                self.patching_config.max_number_of_patches,
                example_sample,
            )
        elif self.patching_config.extraction_type == "fitting":
            patch_parameters["patch_indices"] = self._get_fitting_patching_parameters(
                self.patching_config.patch_size, example_sample
            )
        elif self.patching_config.extraction_type == "overlap":
            (patch_indices, left_padding, right_padding,) = self._get_overlap_patching_parameters(
                self.patching_config.patch_size,
                self.patching_config.overlap_fraction,
                self.patching_config.pad_constant,
                example_sample,
            )
            patch_parameters["patch_indices"] = patch_indices
            patch_parameters["left_padding"] += left_padding
            patch_parameters["right_padding"] += right_padding

        return patch_parameters

    @staticmethod
    def _make_patches(
        image: sitk.Image, patch_parameters: dict, pad_constant: float, patch_size: np.ndarray,
    ) -> list:
        image = SingleSamplePreprocessor._pad_image_from_parameters(
            image,
            patch_parameters["left_padding"],
            patch_parameters["right_padding"],
            pad_constant,
        )

        patches = []
        patch_size = patch_size.tolist()
        patch_filter = sitk.RegionOfInterestImageFilter()
        patch_filter.SetSize(patch_size)

        for i_patch_index in patch_parameters["patch_indices"]:
            patch_filter.SetIndex(i_patch_index.tolist())
            cur_patch = patch_filter.Execute(image)
            patches.append(cur_patch)
        return patches

    @staticmethod
    def _get_random_patching_parameters(
        patch_size: np.ndarray, max_number_of_patches: int, example_sample: ImageSample
    ) -> np.ndarray:
        if max_number_of_patches > 0:
            N_patches = max_number_of_patches
            max_patch_index = example_sample.GetSize() - patch_size
            patch_indices = np.zeros((N_patches, len(max_patch_index)))

            for i_i_max_patch_index, i_max_patch_index in enumerate(max_patch_index):
                patch_indices[:, i_i_max_patch_index] = np.random.randint(
                    0, i_max_patch_index, size=N_patches
                )
            patch_indices = np.floor(patch_indices).astype(np.int)
        else:
            raise ValueError("If extraction is random, number of patches should be specified!")

        return patch_indices

    @staticmethod
    def _get_fitting_patching_parameters(
        patch_size: np.ndarray, example_sample: sitk.Image
    ) -> np.ndarray:
        # sample_size = sample_array.shape
        sample_size = np.asarray(example_sample.GetSize())

        patches_per_dim = np.floor(sample_size / patch_size)
        per_dim_patch_indice_number = []
        for i_patches_per_dim in patches_per_dim:
            per_dim_patch_indice_number.append(range(int(i_patches_per_dim)))
        patch_indice_numbers = np.asarray(list(itertools.product(*per_dim_patch_indice_number)))
        # Spacing is determined by the patch size, and the possible missed voxels
        # because patches dont fit perfectly
        between_patch_spacing = patch_size + np.mod(sample_size, patch_size)/(patches_per_dim -1)
        patch_indices = patch_indice_numbers * between_patch_spacing
        patch_indices = np.floor(patch_indices).astype(np.int)
        return patch_indices

    @staticmethod
    def _get_overlap_patching_parameters(
        patch_size: np.ndarray,
        overlap_fraction: Union[int, list],
        pad_constant: float,
        example_sample: sitk.Image,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

        if not isinstance(overlap_fraction, list):
            overlap_fraction = [overlap_fraction] * example_sample.GetDimension()

        # Need to flip to convert between sitk and numpy coordinates
        image_size = np.asarray(example_sample.GetSize())

        overlap_fraction = np.asarray(overlap_fraction)
        if 0 <= overlap_fraction[0] < 1:
            stride_size = (patch_size - np.ceil(overlap_fraction * patch_size)).astype(np.int)
        else:
            stride_size = (patch_size - overlap_fraction).astype(np.int)

        # Need to calculate the number of stride steps we can take
        # This formula is base so that the full image is covered as muc has possible
        # and as equally as possible, with possible padding as well
        # The total size extracted with the patches = (patch_size + (N-1)*stride_size)
        # With N the number of patches
        # We need to make sure that this is larger than the total image size to ensure
        # that the whole image is cover, hence we take one extra patch (+2 instead of +1)
        # and floor in case the patches dont fit perfectly because we already take an extra one.
        # If we would have used ceil and +1 this would not work for cases where the patches fit
        # perfectly in the image.
        N_stride_steps = np.floor((image_size - patch_size) / stride_size) + 2
        N_stride_steps = N_stride_steps.astype(np.int)

        # We pad so that the steps fit nicely
        required_size = (N_stride_steps - 1) * stride_size + patch_size
        (
            example_sample,
            left_padding,
            right_padding,
        ) = SingleSamplePreprocessor._pad_image_to_size(example_sample, required_size, pad_constant)

        stride_steps_per_dim = []
        for i_i_N_stride_steps, i_N_stride_steps in enumerate(N_stride_steps):
            stride_steps_per_dim.append(range(i_N_stride_steps))

        stride_steps = np.asarray(list(itertools.product(*stride_steps_per_dim)))
        patch_indices = stride_steps * stride_size
        # Need to invert to correspond with the simpleitk indices
        patch_indices = patch_indices.astype(np.int)

        return patch_indices, left_padding, right_padding

    # ===============================================================
    # Rejecting
    # ===============================================================

    def rejecting(self):
        if not self.sample.has_masks:
            raise ValueError("Sample does not have masks, cannot reject patches!")
        rejection_status = self._get_to_reject_patches(
            self.sample.get_example_mask_patches(), self.rejecting_config.rejection_limit,
        )
        if self.rejecting_config.rejection_as_label:
            accepted_status = np.logical_not(rejection_status)
            accepted_status = accepted_status.astype(np.uint8)
            if self.sample.are_labels_one_hot:
                accepted_status = np.eye(2)[accepted_status].astype(np.uint8)
            accepted_labels = [{"accepted": i_status} for i_status in accepted_status]
            self.sample.add_to_labels(accepted_labels, {"accepted": 2})
        else:
            self.sample.channels = (self._get_accepted_patches, [rejection_status])
            self.sample.masks = (self._get_accepted_patches, [rejection_status])

            return self.sample.number_of_patches > 0

    @staticmethod
    def _get_to_reject_patches(mask: Union[sitk.Image, list], rejection_limit: float) -> list:
        if isinstance(mask, sitk.Image):
            mask = [mask]

        rejection_limit = rejection_limit * np.prod(mask[0].GetSize())

        rejection_status = [
            np.count_nonzero(sitk.GetArrayViewFromImage(i_mask_patch)) < rejection_limit
            for i_mask_patch in mask
        ]

        return rejection_status

    @staticmethod
    def _get_accepted_patches(patches: Union[sitk.Image, list], rejection_status: list) -> list:
        accepted_patches = []

        if isinstance(patches, sitk.Image):
            patches = [patches]

        for i_patch, is_rejected in zip(patches, rejection_status):
            if not is_rejected:
                accepted_patches.append(i_patch)
        return accepted_patches

    # ===============================================================
    # Bias field correcting
    # ===============================================================
    def bias_field_correcting(self):
        bias_field_corrector = sitk.N4BiasFieldCorrectionImageFilter()
        if self.bias_field_correcting_config.mask is not None:
            bias_field_corrector.SetUseMaskLabel(True)
            args = [self.bias_field_correcting_config.mask]
        else:
            bias_field_corrector.SetUseMaskLabel(False)
            args = []
        self.sample.channels = (bias_field_corrector.Execute, args)

    # ===============================================================
    # Saving
    # ===============================================================

    @staticmethod
    def _convert_sitk_arrays_to_numpy(images: list):
        N_images = len(images)
        if N_images > 0:
            image_size = images[0].GetSize()
            dtypes = []
            for i_image in images:
                dtypes.append(i_image.GetPixelID())
            # Simpleitk dtypes are ints in increasing order
            # Thus we can get the max and it will it be appropiate everything
            sitk_dtype = np.max(dtypes)
            np_dtype = ImageSample.get_numpy_type_from_sitk_type(sitk_dtype)
            np_array = np.empty((*image_size, N_images), dtype=np_dtype)
            for i_i_image, i_image in enumerate(images):
                np_array[..., i_i_image] = np.transpose(sitk.GetArrayFromImage(i_image))
        else:
            np_array = None
        return np_array

    def _patch_to_data_structure(
        self, patch_channels: list, patch_masks: list, patch_labels: list
    ) -> dict:
        N_channels = len(patch_channels)
        patch_channels = self._convert_sitk_arrays_to_numpy(patch_channels)
        if patch_masks is not None:
            N_masks = len(patch_masks)
            patch_masks = self._convert_sitk_arrays_to_numpy(patch_masks)
        else:
            N_masks = 0

        if self.saving_config.impute_missing_channels:
            patch_channels = self.channel_imputation(patch_channels)

        if self.saving_config.save_as_float16:
            patch_channels = self.channels_to_float16(patch_channels)

        if self.saving_config.use_mask_as_channel and patch_masks is not None:
            patch_names = self.sample.channel_names + self.sample.mask_names
            patches = np.concatenate((patch_channels, patch_masks), axis=-1)
            N_patches = N_channels + N_masks
        else:
            patch_names = self.sample.channel_names
            patches = patch_channels
            N_patches = N_channels

        data_structure = {
            self.saving_config.sample_npz_keyword: {},
            self.saving_config.label_npz_keyword: {},
        }
        if self.saving_config.named_channels:
            data_structure[self.saving_config.sample_npz_keyword] = dict(
                zip(patch_names, np.split(patches, N_patches, axis=-1))
            )
        else:
            data_structure[self.saving_config.sample_npz_keyword] = {
                self.saving_config.sample_npz_keyword: patches
            }

        if self.saving_config.use_mask_as_label:
            if not self.sample.has_masks:
                raise ValueError(
                    "You request to use masks as labels, but no masks were found in the sample!"
                )
            if self.sample.are_labels_one_hot:
                patch_masks = np.squeeze(patch_masks)

                # for i_i_patch_mask, i_patch_mask in enumerate(patch_masks):
                patch_masks = np.eye(self.saving_config.mask_channels)[patch_masks].astype(np.uint8)
            if self.saving_config.named_channels:
                data_structure[self.saving_config.label_npz_keyword] = dict(
                    zip(self.sample.mask_names, np.split(patch_masks, N_masks, axis=-1),)
                )
            elif len(patch_labels) > 0:
                # If we have other labels as well we ensure that we
                # Give a name to the mask, as we have multi-outputs
                data_structure[self.saving_config.label_npz_keyword] = {
                    self.sample.mask_keyword: patch_masks
                }

            else:
                data_structure[self.saving_config.label_npz_keyword] = {
                    self.saving_config.label_npz_keyword: patch_masks
                }

        if len(patch_labels) > 0:

            if self.sample.are_labels_one_hot:
                for i_key, i_value in patch_labels.items():
                    patch_labels[i_key] = np.asarray(i_value).astype(np.int8)
            if self.saving_config.combine_labels:
                label_keys = self.saving_config.label_npz_keyword
                labels = [
                    value for key, value in sorted(patch_labels.items(), key=lambda item: item[0])
                ]
                patch_labels = {self.saving_config.label_npz_keyword: np.asarray(labels)}
            else:
                label_keys = list(patch_labels.keys())
            if len(label_keys) > 1 or self.saving_config.use_mask_as_label:
                data_structure[self.saving_config.label_npz_keyword].update(patch_labels)
            else:
                data_structure[self.saving_config.label_npz_keyword] = {
                    self.saving_config.label_npz_keyword: patch_labels[label_keys[0]]
                }
        return data_structure

    def _get_number_of_classes(self, data_structure: dict):
        labels = data_structure[self.saving_config.label_npz_keyword]
        label_classes = self.sample.number_of_label_classes
        N_labels = len(labels)
        if N_labels == 0:
            number_of_classes = None
        elif N_labels == 1:
            label_key = list(labels.keys())[0]
            if label_classes != {}:
                # we get the label from the sample
                N_classes = list(label_classes.values())[0]
            else:
                # It is a mask, so we get the unique number of classes
                # in the mask
                N_classes = int(len(np.unique(labels[label_key])))
            number_of_classes = {label_key: N_classes}
        else:
            number_of_classes = {}
            for i_key, i_value in labels.items():
                if i_key in label_classes:
                    number_of_classes[i_key] = label_classes[i_key]
                else:
                    number_of_classes[i_key] = int(len(np.unique(i_value)))

        return number_of_classes

    def _write_to_h5py(
        self, filename: str, data_structure: dict, number_of_classes: dict, metadata: dict
    ):
        with h5py.File(filename, "w") as h5f:
            h5f.attrs["sample_name"] = self.sample.sample_name
            sample_group = h5f.create_group(self.saving_config.sample_npz_keyword)
            for i_sample_name, i_sample in data_structure[
                self.saving_config.sample_npz_keyword
            ].items():
                sample_ds = sample_group.create_dataset(i_sample_name, data=i_sample)
                sample_ds.attrs["N_channels"] = np.asarray(i_sample.shape[-1]).astype(np.uint16)
                sample_ds.attrs["size"] = np.asarray(metadata["patch_size"]).astype(np.uint16)
                sample_ds.attrs["dimensionality"] = np.asarray(len(i_sample.shape[0:-1])).astype(
                    np.uint8
                )
                sample_ds.attrs["origin"] = np.asarray(metadata["patch_origin"]).astype(np.float32)
                sample_ds.attrs["index"] = np.asarray(metadata["patch_index"]).astype(np.int32)
                sample_ds.attrs["direction"] = np.asarray(metadata["patch_direction"]).astype(
                    np.float32
                )
                sample_ds.attrs["spacing"] = np.asarray(metadata["patch_spacing"]).astype(
                    np.float32
                )
                sample_ds.attrs["original_size"] = np.asarray(
                    self.sample.original_metadata["image_size"]
                ).astype(np.uint32)
                sample_ds.attrs["original_origin"] = self.sample.original_metadata["image_origin"]
                sample_ds.attrs["original_direction"] = self.sample.original_metadata[
                    "image_direction"
                ]
                sample_ds.attrs["original_spacing"] = self.sample.original_metadata["image_spacing"]

            label_group = h5f.create_group(self.saving_config.label_npz_keyword)
            for i_label_name, i_label in data_structure[
                self.saving_config.label_npz_keyword
            ].items():
                label_ds = label_group.create_dataset(i_label_name, data=i_label)
                label_ds.attrs["N_classes"] = np.asarray(number_of_classes[i_label_name]).astype(
                    np.uint16
                )
                label_ds.attrs["one_hot"] = self.sample.are_labels_one_hot

    def channel_imputation(self, sample_channels):
        sample_channel_names = np.asarray(self.sample.channel_names)
        expected_channel_names = np.asarray(sorted(self.saving_config.channel_names))
        channels_dtype = sample_channels[0].dtype

        imputed_sample_channels = np.zeros(
            sample_channels.shape[0:-1] + (len(expected_channel_names),),
            dtype=sample_channels[0].dtype,
        )

        for i_i, i_expected_channel_name in enumerate(expected_channel_names):
            if i_expected_channel_name in sample_channel_names:
                channel_location = np.squeeze(
                    np.argwhere(sample_channel_names == i_expected_channel_name)
                )
                imputed_sample_channels[..., i_i] = sample_channels[..., channel_location]

        return imputed_sample_channels

    def channels_to_float16(self, sample_channels):
        channels_float16 = sample_channels.astype(np.float16)
        machine_epsilon = np.finfo(sample_channels.dtype).eps
        to_compare = np.copy(sample_channels)

        to_compare[np.abs(to_compare) <= np.finfo(np.float16).eps] = 0

        percentage_diff = np.abs(channels_float16 - to_compare) / (to_compare) * 100.0
        percentage_diff = np.nan_to_num(percentage_diff, posinf=0)
        largest_diff = np.amax(percentage_diff)

        print("largest diff is:")
        print(largest_diff)

        if largest_diff <= self.saving_config.float16_percentage_diff:
            out_channel = channels_float16
        else:
            out_channel = sample_channels

        return out_channel

    def saving(self):
        # First convert to a dict for easy saving in npz format
        sample_channels = self.sample.get_grouped_channels()

        if self.sample.has_masks:
            sample_masks = self.sample.get_grouped_masks()
        else:
            sample_masks = [None] * len(sample_channels)

        sample_metadata = self.sample.metadata
        sample_labels = self.sample.labels

        data_structure_patches = []
        for i_channel_patch, i_mask_patch, i_label_patch in zip(
            sample_channels, sample_masks, sample_labels
        ):
            data_structure_patches.append(
                self._patch_to_data_structure(i_channel_patch, i_mask_patch, i_label_patch)
            )

        if data_structure_patches:
            number_of_classes = self._get_number_of_classes(data_structure_patches[0])
        else:
            number_of_classes = None

        if self.output_directory is not None:
            output_directory = os.path.join(self.output_directory, self.saving_config.out_dir_name)
        else:
            output_directory = self.saving_config.out_dir_name

        IO_utils.create_directory(output_directory)

        default_patch_name = self.sample.sample_name + self._PATCH_SEPARATOR
        if len(data_structure_patches) > 0:
            N_digits = int(np.log10(len(data_structure_patches))) + 1
        else:
            N_digits = 0

        file_names = []

        for (i_i_patch, i_patch), i_metadata in zip(
            enumerate(data_structure_patches), self.sample.metadata
        ):
            out_patch_name = (
                default_patch_name + str(i_i_patch).zfill(N_digits) + self._HDF5_EXTENSION
            )
            out_patch_file = os.path.join(output_directory, out_patch_name)
            self._write_to_h5py(out_patch_file, i_patch, number_of_classes, i_metadata)
            file_names.append(out_patch_file)

        self.save_names = file_names


class BatchPreprocessor:
    def __init__(self, samples_path: str, output_directory: str, config: dict):
        self.sample_directories = IO_utils.get_subdirectories(samples_path)
        self.config = copy.deepcopy(config)
        self.output_directory = output_directory
        self.general_config = PrognosAIs.IO.Configs.general_config(config)
        self.sample_class = Samples.get_sample_class(self.general_config.sample_type)
        self._CLASS_WEIGHT_FILE = "class_weights.json"

        self.labeling_config = PrognosAIs.IO.Configs.labeling_config(config)
        if self.labeling_config.perform_step and self.labeling_config.label_file is not None:
            self.label_loader = PrognosAIs.IO.LabelParser.LabelLoader(
                self.labeling_config.label_file,
                filter_missing=self.labeling_config.filter_missing,
                missing_value=self.labeling_config.missing_value,
                make_one_hot=self.labeling_config.make_one_hot,
            )
        else:
            self.label_loader = None

    def _run_single_sample(self, sample_directory: str):
        sample = self.sample_class(
            root_path=sample_directory, mask_keyword=self.general_config.mask_keyword
        )
        print(sample.sample_name)

        if self.label_loader is not None:
            # print(sample.sample_name in self.label_loader.get_samples())
            if sample.sample_name in self.label_loader.get_samples():
                sample_label = self.label_loader.get_label_from_sample(sample.sample_name)
                number_of_classes = self.label_loader.get_number_of_classes()
                labels_one_hot = self.label_loader.one_hot_encoded
            else:
                return None, None
        else:
            sample_label = {}
            number_of_classes = {}
            labels_one_hot = self.labeling_config.make_one_hot

        sample.add_to_labels(sample_label, number_of_classes)
        sample.are_labels_one_hot = labels_one_hot

        preprocessor = SingleSamplePreprocessor(
            sample, self.config, output_directory=self.output_directory,
        )
        preprocessor.apply_pipeline()
        return preprocessor.save_names, sample_label

    def start(self):
        N_cpus = np.minimum(IO_utils.get_number_of_cpus(), self.general_config.max_cpus)
        print("Number of cpus:")
        print(N_cpus)
        if N_cpus > 1:
            with Pool(N_cpus) as p:
                sample_info = p.map(self._run_single_sample, self.sample_directories)
        else:
            sample_info  = []
            for i_sample_directory in self.sample_directories:
                sample_info.append(self._run_single_sample(i_sample_directory))

        sample_save_names = [
            i_sample_info[0] for i_sample_info in sample_info if i_sample_info[0] is not None
        ]
        single_save_name = sample_save_names[0][0]
        samples_directory = IO_utils.get_parent_directory(single_save_name)

        sample_labels = [
            i_sample_info[1] for i_sample_info in sample_info if i_sample_info[1] is not None
        ]

        self.N_samples = len(sample_save_names)

        subsets = self.split_into_subsets(sample_save_names, sample_labels)

        for subset_name, subset_samples in subsets.items():
            if subset_samples is not None:
                output_folder = os.path.join(samples_directory, subset_name)
                IO_utils.create_directory(output_folder)
                for i_sample in subset_samples:
                    for i_patch in i_sample:
                        shutil.move(i_patch, output_folder)

        if self.label_loader is not None:
            class_weights = self.label_loader.get_class_weights(json_serializable=True)
            with open(
                os.path.join(samples_directory, self._CLASS_WEIGHT_FILE), "w"
            ) as the_json_file:
                json.dump(class_weights, the_json_file, indent=4)

        return samples_directory

    # ===============================================================
    # Labeling
    # ===============================================================

    @staticmethod
    def _get_number_of_samples_in_subsets(
        train_fraction: float,
        validation_fraction: float,
        test_fraction: float,
        number_of_samples: int,
    ):
        set_fractions = np.asarray([train_fraction, validation_fraction, test_fraction])

        # Normalize the fractions
        set_fractions = set_fractions / np.sum(set_fractions)
        N_samples_in_set = set_fractions * number_of_samples

        fractional_samples, N_samples_in_set = np.modf(N_samples_in_set)

        # All the fractional parts we assign to the train set
        N_samples_in_set[0] += np.sum(fractional_samples)
        N_samples_in_set = N_samples_in_set.astype(np.int)

        (N_train_samples, N_val_samples, N_test_samples) = N_samples_in_set

        return N_train_samples, N_val_samples, N_test_samples

    @staticmethod
    def _get_data_split(samples, test_size, stratification_labels=None) -> dict:
        samples = np.asarray(samples)
        if test_size == 0:
            return samples, None, None

        if stratification_labels is not None:
            splitter = sklearn.model_selection.StratifiedShuffleSplit(
                n_splits=1, test_size=test_size
            )
        else:
            splitter = sklearn.model_selection.ShuffleSplit(n_splits=1, test_size=test_size)

        train_indices, test_indices = next(splitter.split(samples, stratification_labels))
        if stratification_labels is not None:
            stratification_labels = np.asarray(stratification_labels)
            train_strat_labels = stratification_labels[train_indices]
        else:
            train_strat_labels = None

        return samples[train_indices], samples[test_indices], train_strat_labels

    def split_into_subsets(self, samples: list, sample_labels: list) -> dict:
        N_train, N_val, N_test = self._get_number_of_samples_in_subsets(
            self.labeling_config.train_fraction,
            self.labeling_config.validation_fraction,
            self.labeling_config.test_fraction,
            self.N_samples,
        )

        if self.labeling_config.stratify_label_name is not None:
            stratification_labels = []
            for i_sample_label in sample_labels:
                stratification_labels.append(
                    i_sample_label[self.labeling_config.stratify_label_name]
                )
        else:
            stratification_labels = None

        samples, test_samples, stratification_labels = self._get_data_split(
            samples, N_test, stratification_labels
        )
        train_samples, val_samples, _ = self._get_data_split(samples, N_val, stratification_labels)

        subsets = {
            "train": train_samples,
            "test": test_samples,
            "validation": val_samples,
        }

        return subsets

    @classmethod
    def init_from_sys_args(cls, args_in):
        parser = argparse.ArgumentParser(
            description="Pre-process (medical) images for use in a neural network"
        )
        parser.add_argument(
            "-c",
            "--config",
            required=True,
            help="The location of the PrognosAIs config file",
            metavar="configuration file",
            dest="config",
            type=str,
        )
        parser.add_argument(
            "-i",
            "--input",
            required=True,
            help="The input directory where the samples to pre-process are located",
            metavar="Input directory",
            dest="input_dir",
            type=str,
        )
        parser.add_argument(
            "-o",
            "--output",
            required=True,
            help="The output directory to store the pre-processed samples in",
            metavar="Output directory",
            dest="output_dir",
            type=str,
        )

        args = parser.parse_args(args_in)

        config = PrognosAIs.IO.ConfigLoader.ConfigLoader(args.config)

        batch_preprocessor = cls(
            args.input_dir, args.output_dir, config.get_preprocessings_settings()
        )

        return batch_preprocessor


if __name__ == "__main__":
    batch_preprocessor = BatchPreprocessor.init_from_sys_args(sys.argv[1:])
    batch_preprocessor.start()



from __future__ import annotations

import argparse
import copy
import json
import logging
import os
import sys

from typing import Tuple
from typing import Union

import tensorflow as tf
from tensorflow.keras.models import load_model


from tensorflow.keras.mixed_precision import experimental as mixed_precision

import PrognosAIs.Constants
import PrognosAIs.IO.utils as IO_utils
import PrognosAIs.Model.Architectures
import PrognosAIs.Model.Parsers as ModelParsers

from PrognosAIs.IO import ConfigLoader
from PrognosAIs.IO import DataGenerator
from PrognosAIs.Model.Architectures import VGG
from PrognosAIs.Model.Architectures import AlexNet
from PrognosAIs.Model.Architectures import DDSNet
from PrognosAIs.Model.Architectures import DenseNet
from PrognosAIs.Model.Architectures import InceptionNet
from PrognosAIs.Model.Architectures import ResNet
from PrognosAIs.Model.Architectures import UNet


class Trainer:
    """Trainer to be used for training a model."""

    def __init__(
        self,
        config: ConfigLoader.ConfigLoader,
        sample_folder: str,
        output_folder: str,
        tmp_data_folder: str = None,
        save_name: str = None,
    ) -> None:
        """
        Trainer to be used for training a model.

        Args:
            config (ConfigLoader.ConfigLoader): Config to be used
            sample_folder (str): Folder containing the train and validation samples
            output_folder (str): Folder to put the resulting model
            tmp_data_folder (str): Folder to copy samples to and load from. Defaults to None.
            save_name (str): Specify a name to save the model as instead of
                using a automatically generated one. Defaults to None.
        """
        self._model = None
        self._train_data_generator_is_setup = False
        self._validation_data_generator_is_setup = False
        self.cluster_resolver = None
        self.total_memory_used = 0
        self.multiworker = False
        self.worker_index = 0
        self.n_workers = 1
        self.steps_per_epoch = None
        self.validation_steps = None

        self.config = copy.deepcopy(config)
        self.output_folder = os.path.join(output_folder, PrognosAIs.Constants.MODEL_SUBFOLDER)

        self.sample_folder = sample_folder
        self.tmp_data_folder = tmp_data_folder

        IO_utils.setup_logger()
        logging.info(
            "Using configuration file: {config_file}".format(config_file=self.config.config_file),
        )
        logging.info("Loading samples from: {input_dir}".format(input_dir=self.sample_folder))
        logging.info("Putting output in: {output}".format(output=self.output_folder))

        if save_name is None:
            self.save_name = self.config.get_save_name()
        else:
            self.save_name = save_name

        self.model_save_file = ".".join(
            [os.path.join(self.output_folder, self.save_name), PrognosAIs.Constants.HDF5_EXTENSION],
        )

        logging.info("Will save model as: {save_name}".format(save_name=self.model_save_file))

        IO_utils.create_directory(self.output_folder)
        self.config.copy_config(self.output_folder, self.save_name)

        # First thing we need to do is get the precision strategy
        # After this we will clear up the GPUs again, so this needs to be done
        # before tensorflow allocates anything (such as in the distribution
        # strategy)
        self.set_precision_strategy(self.config.get_float_policy())

        self.distribution_strategy = self.get_distribution_strategy()

        self.custom_definitions_file = self.config.get_custom_definitions_file()

        self.class_weights = self.load_class_weights()

        self.validation_ds_dir = os.path.join(
            self.sample_folder, PrognosAIs.Constants.VALIDATION_DS_NAME,
        )
        self.do_validation = os.path.exists(self.validation_ds_dir)

    # ===============================================================
    # Distribution and precision strategies
    # ===============================================================

    @staticmethod
    def set_tf_config(
        cluster_resolver: tf.distribute.cluster_resolver.ClusterResolver, environment: str = None,
    ) -> None:
        """
        Set the TF_CONFIG env variable from the given cluster resolver.

        From https://github.com/tensorflow/tensorflow/issues/37693

        Args:
            cluster_resolver (tf.distribute.cluster_resolver.ClusterResolver): cluster
                resolver to use.
            environment (str): Environment to set in TF_CONFIG. Defaults to None.
        """
        cfg = {
            "cluster": cluster_resolver.cluster_spec().as_dict(),
            "task": {
                "type": cluster_resolver.get_task_info()[0],
                "index": cluster_resolver.get_task_info()[1],
            },
            "rpc_layer": cluster_resolver.rpc_layer,
        }
        if environment:
            cfg["environment"] = environment
        os.environ["TF_CONFIG"] = json.dumps(cfg)
        logging.info(
            "Set up TF config environmentas : {TF_CONFIG}".format(
                TF_CONFIG=os.environ["TF_CONFIG"],
            ),
        )

    def set_precision_strategy(self, float_policy_setting: Union[str, bool]) -> None:
        """
        Set the appropiate precision strategy for GPUs.

        If the GPUs support it a mixed float16 precision will be used
        (see tf.keras.mixe_precision for more information), which reduces the memory overhead
        of the training, while doing computation in float32.
        If GPUs dont support mixed precision, we will try a float16 precision setting.
        If that doesn't work either the normal policy is used.
        If you get NaN values for loss or loss doesn't converge it might be because of the policy.
        Try running the model without a policy setting.

        Args:
            float_policy_setting (float_policy_setting: Union[str, bool]): Which policy to select
                if set to PrognosAIs.Constants.AUTO, we will automatically determine what can be done.
                "mixed" will only consider mixed precision, "float16" only considers float16 policy.
                Set to False to not use a policy
        """
        gpus = IO_utils.get_gpu_devices()
        gpu_supports_mixed_precision = all(
            IO_utils.gpu_supports_mixed_precision(i_gpu) for i_gpu in gpus
        )
        gpu_supports_float16_precision = all(IO_utils.gpu_supports_float16(i_gpu) for i_gpu in gpus)

        mixed_precision_allowed = float_policy_setting in [PrognosAIs.Constants.AUTO, "mixed"]
        float16_precision_allowed = float_policy_setting in [PrognosAIs.Constants.AUTO, "float16"]

        if gpu_supports_mixed_precision:
            logging.info("GPU supports a mixed float16 policy")

        if gpu_supports_float16_precision:
            logging.info("GPU support float16 precision policy")

        if gpu_supports_mixed_precision and mixed_precision_allowed and len(gpus) > 0:
            policy = mixed_precision.Policy("mixed_float16", loss_scale="dynamic")
            mixed_precision.set_policy(policy)
            logging.info(
                (
                    "Using a mixed float16 policy, with compute dtype {cdtype} "
                    "and variable dtype {vdtype}.\n"
                    "Loss scaling applied: {loss_scale}"
                ).format(
                    cdtype=policy.compute_dtype,
                    vdtype=policy.variable_dtype,
                    loss_scale=policy.loss_scale,
                ),
            )
            logging.warning(
                "If model is not converging or loss give NaNs, consider turning of this policy",
            )
        elif gpu_supports_float16_precision and float16_precision_allowed and len(gpus) > 0:
            policy = mixed_precision.Policy("float16", loss_scale="dynamic")
            mixed_precision.set_policy(policy)
            logging.info(
                (
                    "Using a float16 policy, with compute dtype {cdtype} "
                    "and variable dtype {vdtype}.\n"
                    "Loss scaling applied: {loss_scale}"
                ).format(
                    cdtype=policy.compute_dtype,
                    vdtype=policy.variable_dtype,
                    loss_scale=policy.loss_scale,
                ),
            )
            logging.warning(
                "If model is not converging or loss give NaNs, consider turning of this policy",
            )
        else:
            logging.info("No float policy is being used")

    def get_distribution_strategy(self) -> tf.distribute.Strategy:
        """
        Get the appropiate distribution strategy.

        A strategy will be returned that can either distribute the training over
        multiple SLURM nodes, over multi GPUs, train on a single GPU or on a
        single CPU (in that order).

        Returns:
            tf.distribute.Strategy: The distribution strategy to be used in training.
        """
        if IO_utils.get_number_of_slurm_nodes() > 1:
            # We are in a slurm environment for multiworker
            self.cluster_resolver = tf.distribute.cluster_resolver.SlurmClusterResolver()
            self.set_tf_config(self.cluster_resolver)
            strategy = tf.distribute.experimental.MultiWorkerMirroredStrategy(
                cluster_resolver=self.cluster_resolver,
                communication=tf.distribute.experimental.CollectiveCommunication.NCCL,
            )
            self.multiworker = True
            self.worker_index = self.cluster_resolver.get_task_info()[1]
            self.n_workers = (
                len(self.cluster_resolver.get_task_info()[0])
                * self.cluster_resolver.max_tasks_per_node
            )
            logging.info(
                "Using a multi-worker distribution environment with {nodes} nodes and {tasks} tasks per node".format(
                    nodes=len(self.cluster_resolver.get_task_info()[0]),
                    tasks=self.cluster_resolver.max_tasks_per_node,
                ),
            )
        elif IO_utils.get_number_of_gpu_devices() > 1:
            strategy = tf.distribute.MirroredStrategy()
            logging.info(
                "Using a mirrored distribution environment with {gpus} parallel GPUs".format(
                    gpus=strategy.num_replicas_in_sync,
                ),
            )
        elif IO_utils.get_number_of_gpu_devices() == 1:
            gpus = IO_utils.get_gpu_devices()
            gpu_device_name = ":".join(gpus[0].name.split(":")[-2:])
            strategy = tf.distribute.OneDeviceStrategy(gpu_device_name)
            logging.info(
                "Using a single device strategy with {device} as device".format(
                    device=gpu_device_name,
                ),
            )
        else:
            cpus = IO_utils.get_cpu_devices()
            cpu_device_name = ":".join(cpus[0].name.split(":")[-2:])
            strategy = tf.distribute.OneDeviceStrategy(cpu_device_name)
            logging.info(
                "Using a single device strategy with {device} as device".format(
                    device=cpu_device_name,
                ),
            )
        return strategy

    # ===============================================================
    # Set-up of data
    # ===============================================================

    def load_class_weights(self) -> Union[None, dict]:
        """
        Load the class weight from the class weight file.

        Returns:
            Union[None, dict]: Class weights if requested and the class weight file exists,
                otherwise None.
        """
        class_weight_file = os.path.join(self.sample_folder, PrognosAIs.Constants.CLASS_WEIGHT_FILE)
        has_class_weight_file = os.path.exists(class_weight_file)
        class_weights = self.config.get_class_weights()

        if not has_class_weight_file and class_weights is None:
            logging.warning("Class weight file not found, not using class weights")

        if self.config.get_use_class_weights() and (
            class_weights is not None or has_class_weight_file
        ):
            if class_weights is None:
                with open(class_weight_file, "r") as the_class_weight_file:
                    class_weights = json.load(the_class_weight_file)

            out_class_weights = {}

            for i_key, i_value in class_weights.items():
                this_class_weight = {}
                for i_class, i_weight in i_value.items():
                    this_class_weight[int(i_class)] = float(i_weight)

                out_class_weights[i_key] = this_class_weight

            if len(out_class_weights.keys()) == 1:
                out_class_weights = list(out_class_weights.values())[0]
            logging.info(
                "Using the following class weights: {weights}".format(weights=out_class_weights)
            )
        else:
            out_class_weights = None
            logging.info(
                "Class weight file found, but requested to not use class weights, thus not using class weights"
            )

        return out_class_weights

    def move_data_to_temporary_folder(self, data_folder: str) -> str:
        """
        Move the data to a temporary directory before loading.

        Args:
            data_folder (str): The original data folder

        Returns:
            str: Folder to which the data has been moved
        """
        if self.tmp_data_folder is not None and self.config.get_copy_files():
            IO_utils.copy_directory(data_folder, self.tmp_data_folder)
            new_folder = os.path.join(self.tmp_data_folder, IO_utils.get_root_name(data_folder))
            logging.info("Loading data from temporary directory {temp}".format(temp=new_folder))
        else:
            new_folder = data_folder
        return new_folder

    @property
    def train_data_generator(self) -> DataGenerator.HDF5Generator:
        """
        The train data generator to be used in training.

        Returns:
            DataGenerator.HDF5Generator: The train data generator
        """
        if not self._train_data_generator_is_setup:
            logging.info("Setting up train data generator")
            train_ds_folder = os.path.join(self.sample_folder, PrognosAIs.Constants.TRAIN_DS_NAME)
            train_ds_folder = self.move_data_to_temporary_folder(train_ds_folder)
            self._train_data_generator = self.setup_data_generator(train_ds_folder)
            self._train_data_generator_is_setup = True
            if self.multiworker:
                self.steps_per_epoch = self.train_data_generator.steps
        return self._train_data_generator

    @property
    def validation_data_generator(self) -> DataGenerator.HDF5Generator:
        """
        The validation data generator to be used in training.

        Returns:
            DataGenerator.HDF5Generator: The validation data generator
        """
        if not self._validation_data_generator_is_setup and self.do_validation:
            logging.info("Setting up validation data generator")
            validation_ds_folder = self.move_data_to_temporary_folder(self.validation_ds_dir)
            self._validation_data_generator = self.setup_data_generator(validation_ds_folder)
            self._validation_data_generator_is_setup = True
            if self.multiworker:
                self.validation_steps = self.validation_data_generator.steps
        else:
            self._validation_data_generator = None
            self._validation_data_generator_is_setup = True
        return self._validation_data_generator

    def setup_data_generator(self, sample_folder: str) -> DataGenerator.HDF5Generator:
        """
        Set up a data generator for a folder containg train samples.

        Args:
            sample_folder (str): The path to the folder containing the sample files.

        Raises:
            ValueError: If the sample folder does not exist or does not contain any samples.

        Returns:
            DataGenerator.HDF5Generator: Datagenerator of the sample in the sample folder.
        """
        if (
            IO_utils.get_root_name(sample_folder) == PrognosAIs.Constants.VALIDATION_DS_NAME
            and not self.config.get_shuffle_val()
        ):
            # In the validation data we don't shuffle and don't apply data augmentation
            # In that way we can be sure that validation metrics will not depend
            # on random components
            data_augmentation = False
            shuffle = False
            logging.warning("Turned off data augmentation and shuffling for validation data")
        else:
            data_augmentation = self.config.get_do_augmentation()
            shuffle = self.config.get_shuffle()

        batch_size = self.config.get_batch_size() * self.distribution_strategy.num_replicas_in_sync
        if self.distribution_strategy.num_replicas_in_sync > 1:
            logging.info(
                (
                    "Requested batch size {batch}, changed to {global_batch} to work correctly with "
                    "distribution strategy"
                ).format(batch=self.config.get_batch_size(), global_batch=batch_size),
            )

        if os.path.exists(sample_folder) and len(os.listdir(sample_folder)) > 0:
            with self.distribution_strategy.scope():
                # Need to multiply the requested batch size by number of replicas
                # in the distribution strategy to get the total global patch sizes
                # the distribution strategy will then make sure that on each device
                # the batch size is the same as the one requested
                data_generator = DataGenerator.HDF5Generator(
                    sample_folder,
                    batch_size=batch_size,
                    shuffle=shuffle,
                    max_steps=self.config.get_max_steps_per_epoch(),
                )

                # We will try to set-up caching of the dataset in memory
                data_generator.setup_caching(
                    self.config.get_cache_in_memory(), self.total_memory_used,
                )
                if data_generator.cache_in_memory:
                    self.total_memory_used += data_generator.memory_size

                if data_augmentation:
                    data_generator.setup_augmentation(
                        self.config.get_data_augmentation_factor(),
                        self.config.get_data_augmentation_settings(),
                    )

                if self.multiworker:
                    data_generator.repeat = True
                    data_generator.setup_sharding(self.worker_index, self.n_workers)
                    logging.info(
                        "Set up sharding with {workers} workers for worker index {index} for the data generator".format(
                            workers=self.n_workers, index=self.worker_index,
                        ),
                    )

        else:
            raise ValueError(
                "Dataset directory {ds_name} does not exist, cannot create data generator!".format(
                    ds_name=sample_folder,
                ),
            )
        return data_generator

    # ===============================================================
    # Model setup
    # ===============================================================

    @property
    def model(self) -> tf.keras.Model:
        """
        Model to be used in training.

        Returns:
            tf.keras.Model: The model
        """
        if self._model is None:
            self._model = self.setup_model()
            logging.info(
                "Using the following model:\n{model}".format(
                    model=self._model.summary(line_length=120),
                ),
            )
        return self._model

    @staticmethod
    def _get_architecture_name(model_name: str, input_dimensionality: dict) -> Tuple[str, str]:
        """
        Get the full architecture name from the model name and input dimensionality.

        Args:
            model_name (str): Name of the model
            input_dimensionality (dict): Dimensionality of the different inputs

        Returns:
            Tuple[str, str]: Class name of architecture and full achitecture name
        """
        separator = "_"
        architecture_name_parts = model_name.split(separator)
        architecture_class_name = architecture_name_parts[0]

        # We get the model that will fit the maximum dimensionality of our inputs
        max_dimensionality = int(max(input_dimensionality.values()))
        full_architecture_name = "{model_name}_{input_dimensionality}D".format(
            model_name=model_name, input_dimensionality=max_dimensionality,
        )

        return architecture_class_name, full_architecture_name

    def _setup_model(self) -> tf.keras.Model:
        """
        Get the model architecture from the architecture name (not yet compiled).

        Raises:
            ValueError: If architecture is not known

        Returns:
            tf.keras.Model: The loaded architecture
        """
        architecture_class_name, full_architecture_name = self._get_architecture_name(
            self.config.get_model_name(), self.train_data_generator.get_feature_dimensionality(),
        )
        architecture_class = getattr(PrognosAIs.Model.Architectures, architecture_class_name, None)
        if architecture_class is None:
            architecture_class = IO_utils.load_module_from_file(
                self.config.get_custom_definitions_file(),
            )
        architecture = getattr(architecture_class, full_architecture_name, None)
        if architecture is None:
            architecture_class = IO_utils.load_module_from_file(
                self.config.get_custom_definitions_file(),
            )
            architecture = getattr(architecture_class, full_architecture_name, None)

        if architecture is None:
            err_msg = "Could not find requested model {model}!".format(model=full_architecture_name)
            raise ValueError(err_msg)

        return architecture(
            self.train_data_generator.get_feature_shape(),
            self.train_data_generator.get_number_of_classes(),
            model_config=self.config.get_model_settings(),
            # TODO SET FROM CONFIG
            input_data_type=tf.keras.backend.floatx(),
        ).create_model()

    def _load_model(self) -> tf.keras.Model:
        logging.info("Loaded the model")
        return load_model(self.config.get_model_file(), compile=False)

    def setup_model(self) -> tf.keras.Model:
        """
        Set up model to be used during train.

        Returns:
            tf.keras.Model: The compiled model to be trained.
        """
        if self.config.get_use_class_weights_in_losses() and self.class_weights is not None:
            loss_parser = ModelParsers.LossParser(
                self.config.get_loss_settings(), self.class_weights, self.custom_definitions_file,
            )
            self.class_weights = None
            logging.info("Using class weights directly inside losses instead of during fit")
        else:
            loss_parser = ModelParsers.LossParser(
                self.config.get_loss_settings(), module_paths=self.custom_definitions_file,
            )
        model_loss = loss_parser.get_losses()

        loss_weights = self.config.get_loss_weights()

        with self.distribution_strategy.scope():
            # Set up the model and compile the model
            # Using the strategy to make sure everything is distributed properly
            logging.info("Setting up model")
            if self.config.get_resume_training_from_model():
                model = self._load_model()
            else:
                model = self._setup_model()

            logging.info("Setting up optimizer")
            optimizer = ModelParsers.OptimizerParser(
                self.config.get_optimizer_settings(), self.custom_definitions_file,
            ).get_optimizer()

            logging.info("Setting up metrics")
            model_metrics = ModelParsers.MetricParser(
                self.config.get_metric_settings(),
                self.train_data_generator.label_names,
                self.custom_definitions_file,
            ).get_metrics()

            model_message = (
                "The following settings are used to set up the model:"
                "Loss: {loss}"
                "Optimizer: {optimizer}"
                "Metrics: {metrics}"
                "Loss weights: {loss_weights}"
            ).format(
                loss=model_loss,
                optimizer=optimizer,
                metrics=model_metrics,
                loss_weights=loss_weights,
            )
            logging.info(model_message)

            model.compile(
                loss=model_loss,
                optimizer=optimizer,
                metrics=model_metrics,
                loss_weights=loss_weights,
            )
        return model

    def setup_callbacks(self) -> list:
        """
        Set up callbacks to be used during training.

        Returns:
            list: the callbacks
        """
        with self.distribution_strategy.scope():
            logging.info("Setting up callbacks")
            callbacks = ModelParsers.CallbackParser(
                self.config.get_callback_settings(),
                self.output_folder,
                self.custom_definitions_file,
                self.save_name,
            ).get_callbacks()
            logging.info("Using the following callbacks: {callbacks}".format(callbacks=callbacks))

        return callbacks

    # ===============================================================
    # Model training
    # ===============================================================

    def train_model(self) -> str:
        """
        Train the model.

        Returns:
            str: The location where the model has been saved
        """
        with self.distribution_strategy.scope():
            train_data = self.train_data_generator.get_tf_dataset()
            if self.do_validation:
                validation_data = self.validation_data_generator.get_tf_dataset()
            else:
                validation_data = None

        epochs = self.config.get_N_epoch()
        callbacks = self.setup_callbacks()
        logging.info("Starting training")
        logging.debug(
            (
                "Training with following parameters:\n"
                "Train data: {train}\n"
                "Validation data: {val}\n"
                "Epochs: {epoch}\n"
                "Callbacks: {callback}\n"
                "Class weights: {weights}\n"
                "Steps per epoch: {steps}\n"
                "Validation setps: {val_steps}\n"
            ).format(
                train=train_data,
                val=validation_data,
                epoch=epochs,
                callback=callbacks,
                weights=self.class_weights,
                steps=self.steps_per_epoch,
                val_steps=self.validation_steps,
            ),
        )

        self.model.fit(
            train_data,
            validation_data=validation_data,
            epochs=epochs,
            callbacks=callbacks,
            shuffle=False,
            class_weight=self.class_weights,
            verbose=1,
            steps_per_epoch=self.steps_per_epoch,
            validation_steps=self.validation_steps,
        )

        logging.info("Finished training")
        if self.worker_index == 0:
            self.model.save(self.model_save_file)
            logging.info("Model saved to {save_file}".format(save_file=self.model_save_file))
        else:
            # We need to save the model for other workers as well, otherwise
            # We run into errors, however we instantly delete because we dont actually
            # Need the other models
            model_save_file = ".".join(
                [
                    os.path.join(self.output_folder, self.save_name + "_" + str(self.worker_index)),
                    PrognosAIs.Constants.HDF5_EXTENSION,
                ],
            )

            self.model.save(model_save_file)
            os.remove(model_save_file)
        return self.model_save_file

    # ===============================================================
    # External use
    # ===============================================================

    @classmethod
    def init_from_sys_args(cls: Trainer, args_in: list) -> Trainer:
        """
        Initialize a Trainer object from the command line.

        Args:
            args_in (list): Arguments to parse to the trainer

        Returns:
            Trainer: The trainer object
        """
        parser = argparse.ArgumentParser(description="Train a CNN")

        parser.add_argument(
            "-c",
            "--config",
            required=True,
            help="The location of the PrognosAIs config file",
            metavar="configuration file",
            dest="config",
            type=str,
        )

        parser.add_argument(
            "-i",
            "--input",
            required=True,
            help="The input directory where the samples to train on are located",
            metavar="Input directory",
            dest="input_dir",
            type=str,
        )

        parser.add_argument(
            "-o",
            "--output",
            required=True,
            help="The output directory where to store the saved model",
            metavar="Output directory",
            dest="output_dir",
            type=str,
        )
        parser.add_argument(
            "-T",
            "--tmp",
            required=False,
            help="The temporary directory",
            metavar="Temp directory",
            dest="tmp_dir",
            type=str,
            default=None,
        )
        parser.add_argument(
            "-s",
            "--savename",
            required=False,
            help="Save name for model",
            metavar="Save name",
            dest="save_name",
            type=str,
            default=None,
        )

        args = parser.parse_args(args_in)

        config = ConfigLoader.ConfigLoader(args.config)

        return cls(config, args.input_dir, args.output_dir, args.tmp_dir, args.save_name)


if __name__ == "__main__":
    trainer = Trainer.init_from_sys_args(sys.argv[1:])
    trainer.train_model()



import numpy as np
import tensorflow as tf

from tensorflow import keras

# from tensorflow import distribute
# from tensorflow.keras.backend import print_tensor
# from tensorflow.python.ops.losses import util as tf_losses_utils
# from tensorflow.python.framework import dtypes
from tensorflow.keras import layers
from tensorflow.python.keras import backend as K
from tensorflow.python.keras.utils import metrics_utils
from tensorflow.python.ops import math_ops
from tensorflow.python.framework import ops


def concordance_index(y_true, y_pred):
    """
    This function determines the concordance index given two tensorflow tensors

    y_true contains a label to indicate whether events occurred, and time to events
    (or time to right censored data if no event occurred)

    y_pred is beta*x in the cox model
    """
    events_occurred = y_true[:, 0]
    time_to_events = y_true[:, 1]

    # time_to_events = print_tensor(time_to_events, message='time to events: ')

    y_pred = tf.squeeze(y_pred)
    identity_matrix = tf.cast(y_pred[None, :] < y_pred[:, None], tf.float32)

    risk_set = tf.cast(time_to_events[None, :] > time_to_events[:, None], tf.float32)

    risk_set_idenity_matrix = tf.math.multiply(identity_matrix, risk_set)
    correct_prediction = tf.math.reduce_sum(risk_set_idenity_matrix, axis=1)

    correct_prediction_event_set = tf.math.multiply(correct_prediction, events_occurred)

    concordance_index_unscaled = tf.math.reduce_sum(correct_prediction_event_set)

    concordance_index = concordance_index_unscaled / tf.cast(
        tf.math.count_nonzero(identity_matrix), tf.float32
    )

    return concordance_index


class ConcordanceIndex(keras.metrics.Metric):
    def __init__(self, name="ConcordanceIndex", **kwargs):
        super().__init__(name=name, **kwargs)
        self.summed_concordance = self.add_weight(
            name="summed_corcodance", initializer="zeros", dtype=tf.float32
        )
        self.total_samples = self.add_weight(
            name="total_samples", initializer="zeros", dtype=tf.float32
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_pred = tf.squeeze(y_pred)
        events_occurred = tf.squeeze(y_true[:, 0])
        time_to_events = tf.squeeze(y_true[:, 1])
        identity_matrix = tf.cast(y_pred[None, :] < y_pred[:, None], tf.float32)

        risk_set = tf.cast(events_occurred[None, :] > events_occurred[:, None], tf.float32)

        risk_set_idenity_matrix = tf.math.multiply(identity_matrix, risk_set)
        correct_prediction = tf.math.reduce_sum(risk_set_idenity_matrix, axis=1)

        correct_prediction_event_set = tf.math.multiply(correct_prediction, time_to_events)

        concordance_index_unscaled = tf.math.reduce_sum(correct_prediction_event_set)

        concordance_index = concordance_index_unscaled / tf.cast(
            tf.math.count_nonzero(identity_matrix), tf.float32
        )

        self.summed_concordance.assign_add(concordance_index)
        self.total_samples.assign_add(tf.cast(tf.shape(y_true)[0], tf.float32))

    def result(self):
        return tf.math.divide_no_nan(self.summed_concordance, self.total_samples)


class Sensitivity(keras.metrics.Metric):
    def __init__(self, name="Sensitivity_custom", **kwargs):
        super(Sensitivity, self).__init__(name=name, **kwargs)
        self.true_positives = self.add_weight(
            name="true_positives", shape=(1,), initializer="zeros"
        )
        self.true_negatives = self.add_weight(
            name="true_negatives", shape=(1,), initializer="zeros"
        )
        self.false_positives = self.add_weight(
            name="false_positives", shape=(1,), initializer="zeros"
        )
        self.false_negatives = self.add_weight(
            name="false_negatives", shape=(1,), initializer="zeros"
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        # Get the actual labels
        # y_true = tf.squeeze(y_true)
        # y_pred = tf.argmax(tf.squeeze(y_pred), axis=1)

        metrics_utils.update_confusion_matrix_variables(
            {
                metrics_utils.ConfusionMatrix.TRUE_POSITIVES: self.true_positives,
                metrics_utils.ConfusionMatrix.TRUE_NEGATIVES: self.true_negatives,
                metrics_utils.ConfusionMatrix.FALSE_POSITIVES: self.false_positives,
                metrics_utils.ConfusionMatrix.FALSE_NEGATIVES: self.false_negatives,
            },
            y_true,
            y_pred,
            thresholds=[0.5],
            sample_weight=sample_weight,
        )

    def result(self):
        result = math_ops.div_no_nan(
            self.true_positives, self.true_positives + self.false_negatives
        )
        return result[0]

    def reset_states(self):
        K.batch_set_value([(v, np.zeros((1,))) for v in self.variables])


class Specificity(keras.metrics.Metric):
    def __init__(self, name="Specificity_custom", **kwargs):
        super(Specificity, self).__init__(name=name, **kwargs)
        self.true_positives = self.add_weight(
            name="true_positives", shape=(1,), initializer="zeros"
        )
        self.true_negatives = self.add_weight(
            name="true_negatives", shape=(1,), initializer="zeros"
        )
        self.false_positives = self.add_weight(
            name="false_positives", shape=(1,), initializer="zeros"
        )
        self.false_negatives = self.add_weight(
            name="false_negatives", shape=(1,), initializer="zeros"
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        # y_true = tf.squeeze(y_true)
        # y_pred = tf.argmax(tf.squeeze(y_pred), axis=1)

        metrics_utils.update_confusion_matrix_variables(
            {
                metrics_utils.ConfusionMatrix.TRUE_POSITIVES: self.true_positives,
                metrics_utils.ConfusionMatrix.TRUE_NEGATIVES: self.true_negatives,
                metrics_utils.ConfusionMatrix.FALSE_POSITIVES: self.false_positives,
                metrics_utils.ConfusionMatrix.FALSE_NEGATIVES: self.false_negatives,
            },
            y_true,
            y_pred,
            thresholds=[0.5],
            sample_weight=sample_weight,
        )

    def result(self):
        result = math_ops.div_no_nan(
            self.true_negatives, self.true_negatives + self.false_positives
        )
        return result[0]

    def reset_states(self):
        K.batch_set_value([(v, np.zeros((1,))) for v in self.variables])


class MaskedAUC(keras.metrics.AUC):
    def __init__(self, name="MaskedAUC", mask_value=-1, **kwargs):
        if "multi_label" not in kwargs:
            super().__init__(name=name, multi_label=True, **kwargs)
        else:
            super().__init__(name=name, **kwargs)
        self.mask_value = mask_value

    def update_state(self, y_true, y_pred, sample_weight=None):
        # we remove masked samples from y_true and y_pred
        y_pred = tf.clip_by_value(y_pred, 0, 1)
        y_pred = ops.convert_to_tensor(y_pred)
        y_true = math_ops.cast(y_true, y_pred.dtype)

        unmasked_samples = tf.reduce_all(tf.math.not_equal(y_true, self.mask_value), axis=-1)
        unmasked_samples = tf.cast(unmasked_samples, self.dtype)

        weight_correction = unmasked_samples * tf.math.divide_no_nan(
            tf.cast(tf.gather(tf.shape(y_true), 0), self.dtype),
            tf.math.count_nonzero(unmasked_samples, dtype=self.dtype),
        )

        if sample_weight is not None:
            sample_weight *= weight_correction
        else:
            sample_weight = weight_correction

        super().update_state(y_true, y_pred, sample_weight)

    def get_config(self):
        config = super().get_config()
        config["mask_value"] = self.mask_value
        return config


class MaskedCategoricalAccuracy(keras.metrics.CategoricalAccuracy):
    def __init__(self, name="MaskedCategoricalAccuracy", mask_value=-1, **kwargs):
        super().__init__(name=name, **kwargs)
        self.mask_value = mask_value

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_pred = ops.convert_to_tensor(y_pred)
        y_true = math_ops.cast(y_true, y_pred.dtype)

        masked_samples = tf.reduce_all(tf.math.not_equal(y_true, self.mask_value), axis=-1)
        masked_samples = math_ops.cast(masked_samples, y_true.dtype)
        masked_samples.set_shape([None])

        y_true = tf.boolean_mask(y_true, masked_samples)
        y_pred = tf.boolean_mask(y_pred, masked_samples)

        if sample_weight is not None:
            sample_weight = tf.boolean_mask(sample_weight, masked_samples)

        super().update_state(y_true, y_pred, sample_weight)

    def get_config(self):
        config = super().get_config()
        config["mask_value"] = self.mask_value
        return config


class DICE(keras.metrics.Metric):
    def __init__(self, name="dice_coefficient", foreground_only=True, **kwargs):
        super().__init__(name=name, **kwargs)
        self.summed_dice = self.add_weight(
            name="summed_dice", initializer="zeros", dtype=tf.float32
        )
        self.total_samples = self.add_weight(
            name="total_samples", initializer="zeros", dtype=tf.float32
        )
        self.foreground_only = foreground_only

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_pred = ops.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)
        y_pred = tf.math.round(y_pred)

        to_reduce_axis = tf.range(1, tf.rank(y_true) - 1, 1)
        numerator = tf.math.reduce_sum(y_true * y_pred, axis=to_reduce_axis)
        denominator = tf.math.reduce_sum(y_true + y_pred, axis=to_reduce_axis)

        # We first calculate the loss (1 -dice, which adds he denominator in te numerato)
        # Because thn using divide_no_nan give 0 for the loss if no
        dice_per_class = 1.0 - tf.math.divide_no_nan(denominator - 2.0 * numerator, denominator)

        if self.foreground_only:
            dice = tf.math.reduce_sum(dice_per_class[..., 1:])
        else:
            dice = tf.math.reduce_sum(tf.math.reduce_mean(dice_per_class))

        self.summed_dice.assign_add(dice)
        self.total_samples.assign_add(tf.cast(tf.shape(y_true)[0], tf.float32))

    def result(self):
        return tf.math.divide_no_nan(self.summed_dice, self.total_samples)

    def get_config(self):
        config = super().get_config()
        config["foreground_only"] = self.foreground_only

        return config


class MaskedSpecificity(keras.metrics.Metric):
    def __init__(self, name="masked_specificity", mask_value=-1, **kwargs):
        super().__init__(name=name, **kwargs)
        self.true_positives = self.add_weight(
            name="true_positives", shape=(1,), initializer="zeros"
        )
        self.true_negatives = self.add_weight(
            name="true_negatives", shape=(1,), initializer="zeros"
        )
        self.false_positives = self.add_weight(
            name="false_positives", shape=(1,), initializer="zeros"
        )
        self.false_negatives = self.add_weight(
            name="false_negatives", shape=(1,), initializer="zeros"
        )
        self.mask_value = mask_value

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_pred = ops.convert_to_tensor(y_pred)
        y_true = math_ops.cast(y_true, y_pred.dtype)

        masked_samples = tf.reduce_all(tf.math.not_equal(y_true, self.mask_value), axis=-1)
        masked_samples = math_ops.cast(masked_samples, y_true.dtype)
        masked_samples.set_shape([None])

        y_true = tf.boolean_mask(y_true, masked_samples)
        y_pred = tf.boolean_mask(y_pred, masked_samples)

        y_true = tf.math.argmax(y_true, axis=-1)
        y_pred = tf.math.argmax(y_pred, axis=-1)

        metrics_utils.update_confusion_matrix_variables(
            {
                metrics_utils.ConfusionMatrix.TRUE_POSITIVES: self.true_positives,
                metrics_utils.ConfusionMatrix.TRUE_NEGATIVES: self.true_negatives,
                metrics_utils.ConfusionMatrix.FALSE_POSITIVES: self.false_positives,
                metrics_utils.ConfusionMatrix.FALSE_NEGATIVES: self.false_negatives,
            },
            y_true,
            y_pred,
            thresholds=[0.5],
            sample_weight=sample_weight,
        )

    def result(self):
        result = math_ops.div_no_nan(
            self.true_negatives, self.true_negatives + self.false_positives
        )
        return result[0]

    def reset_states(self):
        K.batch_set_value([(v, np.zeros((1,))) for v in self.variables])


class MaskedSensitivity(keras.metrics.Metric):
    def __init__(self, name="masked_sensitivity", mask_value=-1, **kwargs):
        super().__init__(name=name, **kwargs)
        self.true_positives = self.add_weight(
            name="true_positives", shape=(1,), initializer="zeros"
        )
        self.true_negatives = self.add_weight(
            name="true_negatives", shape=(1,), initializer="zeros"
        )
        self.false_positives = self.add_weight(
            name="false_positives", shape=(1,), initializer="zeros"
        )
        self.false_negatives = self.add_weight(
            name="false_negatives", shape=(1,), initializer="zeros"
        )
        self.mask_value = mask_value

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_pred = ops.convert_to_tensor(y_pred)
        y_true = math_ops.cast(y_true, y_pred.dtype)

        masked_samples = tf.reduce_all(tf.math.not_equal(y_true, self.mask_value), axis=-1)
        masked_samples = math_ops.cast(masked_samples, y_true.dtype)
        masked_samples.set_shape([None])

        y_true = tf.boolean_mask(y_true, masked_samples)
        y_pred = tf.boolean_mask(y_pred, masked_samples)

        y_true = tf.math.argmax(y_true, axis=-1)
        y_pred = tf.math.argmax(y_pred, axis=-1)

        if sample_weight is not None:
            sample_weight = tf.boolean_mask(sample_weight, masked_samples)

        metrics_utils.update_confusion_matrix_variables(
            {
                metrics_utils.ConfusionMatrix.TRUE_POSITIVES: self.true_positives,
                metrics_utils.ConfusionMatrix.TRUE_NEGATIVES: self.true_negatives,
                metrics_utils.ConfusionMatrix.FALSE_POSITIVES: self.false_positives,
                metrics_utils.ConfusionMatrix.FALSE_NEGATIVES: self.false_negatives,
            },
            y_true,
            y_pred,
            thresholds=[0.5],
            sample_weight=sample_weight,
        )

    def result(self):
        result = math_ops.div_no_nan(
            self.true_positives, self.true_positives + self.false_negatives
        )
        return result[0]

    def reset_states(self):
        K.batch_set_value([(v, np.zeros((1,))) for v in self.variables])



from abc import ABC
from abc import abstractmethod
from typing import Union

import numpy as np
import tensorflow

from tensorflow.keras import Input
from tensorflow.keras.backend import int_shape
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Dropout


class NetworkArchitecture(ABC):
    def __init__(
        self,
        input_shapes: dict,
        output_info: dict,
        input_data_type="float32",
        output_data_type="float32",
        model_config: dict = {},
    ):
        self.input_data_type = input_data_type
        self.output_data_type = output_data_type
        self.input_shapes = input_shapes
        self.output_info = output_info
        self.model_config = model_config
        # self.inputs = self.make_inputs(input_shapes, input_data_type)
        # self.outputs = self.make_outputs(output_info, output_data_type)

    def make_inputs(
        self, input_shapes: dict, input_dtype: str, squeeze_inputs: bool = True
    ) -> Union[dict, Input]:
        inputs = {}
        for i_input_name, i_input_shape in input_shapes.items():
            # TODO pass correct input type according to policy
            inputs[i_input_name] = Input(shape=i_input_shape, name=i_input_name)

        if squeeze_inputs and len(inputs) == 1:
            inputs = list(inputs.values())[0]

        return inputs

    @staticmethod
    def check_minimum_input_size(input_layer: Input, minimum_input_size: np.ndarray):
        input_shape = int_shape(input_layer)
        if len(input_shape) - 1 == len(minimum_input_size):
            err_msg = (
                "It seems like you have forgotten to include a"
                " (potentially empty) channel dimension as the last dimension.\n"
                "Please fix this and run the model again."
            )
            raise ValueError(err_msg)
        input_shape = input_shape[1:-1]
        minimum_input_size = np.asarray(minimum_input_size)

        if any(input_shape < minimum_input_size):
            min_inputs = [str(i_input_size) for i_input_size in minimum_input_size]
            min_input_format = " x ".join(min_inputs)

            cur_inputs = [str(i_input_shape) for i_input_shape in input_shape]
            cur_input_format = " x ".join(cur_inputs)

            err_msg = (
                "Minimum input size for this model is: {}\n"
                "Your input size is: {}\n"
                "Please fix your input"
            ).format(min_input_format, cur_input_format,)
            raise ValueError(err_msg)

    @staticmethod
    def get_corrected_stride_size(
        layer: tensorflow.keras.layers, stride_size: list, conv_size: list
    ):
        """
        Ensure that the stride is never bigger than the actual input
        In this way any network can keep working, indepedent of size
        """

        input_shape = int_shape(layer)[1:-1]
        stride_size = np.asarray(stride_size)
        conv_size = np.asarray(conv_size)

        stride_total_size = stride_size * conv_size

        stride_size = np.floor(np.minimum(input_shape, stride_total_size) / conv_size)

        # Ensure that stride_size is at least one
        stride_size = np.maximum(stride_size, np.ones(len(stride_size)))
        stride_size = stride_size.astype(np.int).tolist()

        return stride_size

    def make_dropout_layer(self, layer):
        if "dropout" in self.model_config and self.model_config["dropout"] > 0:
            out_layer = Dropout(self.model_config["dropout"])(layer)
        else:
            out_layer = layer
        return out_layer

    @abstractmethod
    def make_outputs(self, output_info: dict, output_data_type: str) -> tensorflow.keras.layers:
        """Make the outputs"""

    @abstractmethod
    def create_model(self):
        """Here the code to create the actual model"""


class ClassificationNetworkArchitecture(NetworkArchitecture):
    def __init__(
        self,
        input_shapes: dict,
        output_info: dict,
        input_data_type="float32",
        output_data_type="float32",
        model_config={},
    ):

        super().__init__(input_shapes, output_info, input_data_type, output_data_type, model_config)
        self.inputs = self.make_inputs(input_shapes, input_data_type)
        self.outputs = self.make_outputs(output_info, output_data_type)

    @staticmethod
    def make_outputs(
        output_info: dict,
        output_data_type: str,
        activation_type: str = "softmax",
        squeeze_outputs: bool = True,
    ) -> dict:
        outputs = {}
        for i_output_name, i_output_classes in output_info.items():
            outputs[i_output_name] = Dense(
                i_output_classes, name=i_output_name, activation=activation_type, dtype="float32",
            )

        if squeeze_outputs and len(outputs) == 1:
            outputs = list(outputs.values())[0]

        return outputs



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['PrognosAIs',
 'PrognosAIs.IO',
 'PrognosAIs.Model',
 'PrognosAIs.Model.Architectures',
 'PrognosAIs.Preprocessing']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'SimpleITK>=1.2.4,<3.0.0',
 'h5py>=2.10.0,<3.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'natsort>=7.0.1,<8.0.0',
 'numba>=0.49.1,<0.53.0',
 'numpy>=1.18,<2.0',
 'pandas>=1.0.3,<2.0.0',
 'psutil>=5.7.0,<6.0.0',
 'scikit-learn>=0.23,<0.25',
 'slurmpie>=0.4.0,<0.5.0',
 'tensorboard_plugin_profile>=2.2.0,<3.0.0',
 'tensorflow-addons>=0.11',
 'tensorflow-io>=0.15',
 'tensorflow>=2.2.0']

setup_kwargs = {
    'name': 'prognosais',
    'version': '0.2.7',
    'description': 'Tool to quickly and easily train CNNs for medical imaging tasks',
    'long_description': '[![Documentation Status](https://readthedocs.org/projects/prognosais/badge/?version=latest)](https://prognosais.readthedocs.io/en/latest/?badge=latest) [![codecov](https://codecov.io/gh/Svdvoort/prognosais/branch/master/graph/badge.svg?token=HTHVINR6Y8)](https://codecov.io/gh/Svdvoort/prognosais) [![Python test](https://github.com/Svdvoort/prognosais/workflows/Python%20test/badge.svg)](https://github.com/Svdvoort/prognosais/actions?query=workflow%3A%22Python+test%22)[![CodeFactor](https://www.codefactor.io/repository/github/svdvoort/prognosais/badge)](https://www.codefactor.io/repository/github/svdvoort/prognosais)[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![Dependabot](https://flat.badgen.net/dependabot/Svdvoort/prognosais?icon=dependabot)](https://github.com/Svdvoort/prognosais/pulls?q=is%3Aopen+is%3Apr+label%3Adependencies)\n\n# prognosais\n\nprognosais is a tool to quickly prototype some CNNs for medical classification and segmentations tasks.\nPlease read the documentation: https://prognosais.readthedocs.io/en/latest/?badge=latest\n',
    'author': 'Sebastian van der Voort',
    'author_email': 'svoort25@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Svdvoort/prognosais',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.9',
}


setup(**setup_kwargs)



import datetime
import hashlib
import os
import random
import re
import shutil

import PrognosAIs.IO.utils as utils
import yaml


# TODO make sure in testing that all floats/ number are actually loaded as such
class ConfigLoader:
    def __init__(self, config_file):
        self.config_file = config_file
        loader = yaml.SafeLoader
        # We add the following to safely load scientific notation as floats
        loader.add_implicit_resolver(
            "tag:yaml.org,2002:float",
            re.compile(
                """^(?:
                       [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
                       |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
                       |\\.[0-9_]+(?:[eE][-+][0-9]+)?
                       |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
                       |[-+]?\\.(?:inf|Inf|INF)
                       |\\.(?:nan|NaN|NAN))$""",
                re.X,
            ),
            list("-+0123456789."),
        )

        with open(config_file, "r") as ymlfile:
            self.cfg = yaml.load(ymlfile, Loader=loader)

        if "original_location" not in self.cfg:
            self.cfg["original_location"] = utils.get_parent_directory(self.config_file)

    def copy_config(self, output_folder, save_name=None):
        if save_name is not None:
            new_config_location = os.path.join(
                output_folder, "config_{savename}.yaml".format(savename=save_name)
            )
        else:
            new_config_location = os.path.join(output_folder, "config.yaml")

        with open(new_config_location, "w") as f:
            yaml.dump(self.cfg, f, default_flow_style=False)

        return new_config_location

    def get_config_file(self):
        return self.config_file

    def get_data_folder(self):
        return self.cfg["preprocessing"]["data_folder"]

    def get_test_data_folder(self):
        return self.cfg["testing"]["data_folder"]

    def get_resample_images(self):
        return self.cfg["preprocessing"]["resample_images"]

    def get_make_patches(self):
        return self.cfg["preprocessing"]["make_patches"]

    def get_patch_size(self):
        return self.cfg["preprocessing"]["patch_size"]

    def get_reject_patches(self):
        return self.cfg["preprocessing"]["reject_patches"]

    def get_min_patch_voxels(self):
        return self.cfg["preprocessing"]["min_patch_voxels"]

    def get_extra_input_file(self):
        return self.cfg["preprocessing"]["extra_input_file"]

    def get_mask_file(self):
        return self.cfg["preprocessing"]["mask_file"]

    def get_fsl_reorient_bin(self):
        return self.cfg["program_paths"]["fslreorient2std_bin"]

    def get_fsl_val_bin(self):
        return self.cfg["program_paths"]["fslval_bin"]

    def get_label_file(self):
        return self.cfg["preprocessing"]["label_file"]

    def get_test_label_file(self):
        return self.cfg["testing"]["label_file"]

    def get_dataset_distribution(self):
        return self.cfg["preprocessing"]["dataset_distribution"]

    def get_resample_size(self):
        return self.cfg["preprocessing"]["resample_image_size"]

    def get_random_state(self):
        random_state = self.cfg["preprocessing"]["random_state"]
        if random_state == -1:
            random_state = random.randint(1, 5000000)
        return random_state

    def get_stratify_index(self):
        stratify_index = self.cfg["preprocessing"]["stratify_index"]
        if stratify_index == "None":
            stratify_index = None
        return stratify_index

    def get_multi_channels_patches(self):
        return self.cfg["preprocessing"]["multi_channel_patches"]

    def get_N_max_patches(self):
        return self.cfg["preprocessing"]["N_max_patches"]

    def get_use_mask_as_channel(self):
        return self.cfg["preprocessing"]["use_mask_as_channel"]

    def get_use_mask_as_label(self):
        return self.cfg["preprocessing"]["use_mask_as_label"]

    def get_keep_rejected_patches(self):
        return self.cfg["preprocessing"]["keep_rejected_patches"]

    def get_mask_keyword(self):
        return self.cfg["preprocessing"]["mask_keyword"]

    def get_center_patch_around_mask(self):
        return self.cfg["preprocessing"]["center_patch_around_mask"]

    def get_rescale_mask_intensity(self):
        return self.cfg["preprocessing"]["rescale_mask_intensity"]

    def get_model_name(self):
        return self.cfg["model"]["architecture"]["model_name"]

    def get_model_settings(self):
        if "settings" in self.cfg["model"]["architecture"]:
            model_settings = self.cfg["model"]["architecture"]["settings"]
            if model_settings is None:
                model_settings = {}
        else:
            model_settings = {}
        return model_settings

    def get_cache_in_memory(self):
        if "cache_in_memory" in self.cfg["training"]:
            cache_in_memory = self.cfg["training"]["cache_in_memory"]
        else:
            cache_in_memory = False
        return cache_in_memory

    def get_float_policy(self):
        if "float_policy" in self.cfg["training"]:
            float_policy = self.cfg["training"]["float_policy"]
        else:
            float_policy = False
        return float_policy

    def get_gpu_workers(self):
        return self.cfg["training"]["N_workers"]

    def get_shuffle(self):
        return self.cfg["training"]["shuffle"]

    def get_data_augmentation(self):
        return self.cfg["training"]["data_augmentation"]

    def get_data_augmentation_factor(self):
        return self.cfg["training"]["augmentation_factor"]

    def get_batch_size(self):
        return self.cfg["model"]["architecture"]["batch_size"]

    def get_N_epoch(self):
        return self.cfg["model"]["architecture"]["N_epoch"]

    def get_N_classes(self):
        return self.cfg["model"]["architecture"]["N_output"]

    def get_make_one_hot(self):
        return self.cfg["training"]["make_one_hot"]

    def get_filter_missing(self):
        return self.cfg["training"]["filter_missing"]

    def get_dtype(self):
        return self.cfg["network"]["dtype"]

    def get_copy_files(self):
        return self.cfg["training"]["copy_files"]

    def get_test_model_file(self):
        return self.cfg["testing"]["model_file"]

    def get_max_steps_per_epoch(self):
        return self.cfg["training"]["max_steps_per_epoch"]

    def get_training_multi_processing(self):
        return self.cfg["training"]["multi_processing"]

    def get_float16_epsilon(self):
        if "float16_epsilon" in self.cfg["training"]:
            return float(self.cfg["training"]["float16_epsilon"])
        else:
            return 1e-4

    def get_save_name(self):
        time_string = datetime.datetime.now().strftime("%Y_%m_%d_%H:%M:%S")
        return time_string

    def get_image_size(self):
        if self.get_resample_images():
            return self.cfg["preprocessing"]["resample_image_size"]
        elif self.get_make_patches():
            return self.cfg["preprocessing"]["patch_size"]
        return None

    def get_size_string(self):
        size = self.get_image_size()
        if size is not None:
            size_string = [str(size[i]) + "x" for i in range(len(size))]
            size_string = "".join(size_string)
            size_string = size_string[0:-1]
        else:
            size_string = None
        return size_string

    def get_processed_samples_folder(self):
        base_dir = utils.get_parent_directory(self.get_data_folder())
        processed_samples_folder = os.path.join(base_dir, "NPZ_samples_" + self.get_size_string())

        return processed_samples_folder

    def get_output_folder(self):
        return self.cfg["general"]["output_folder"]

    def get_input_folder(self):
        return self.cfg["general"]["input_folder"]

    def get_cluster_setting(self):
        if "cluster_type" in self.cfg["general"]:
            return self.cfg["general"]["cluster_type"]
        else:
            return None

    def get_specific_output_folder(self):
        hasher = hashlib.sha512()
        with open(self.config_file, "rb") as afile:
            buf = afile.read()
            hasher.update(buf)
        hash_string = hasher.hexdigest()
        output_specifications = [
            self.get_model_name(),
            self.get_save_name(),
            hash_string,
        ]
        output_string = "_".join(output_specifications)
        # specific_output_folder = os.path.join(self.get_output_folder(), output_string)

        # utils.create_directory(specific_output_folder)

        return output_string

    def get_N_jobs(self):
        return self.cfg["cluster"]["N_jobs"]

    def get_cluster_type(self):
        return self.cfg["cluster"]["type"]

    def get_use_labels_from_rejection(self):
        return self.cfg["preprocessing"]["labels_from_rejection"]

    def get_use_class_weights(self):
        return self.cfg["training"]["use_class_weights"]

    def get_use_class_weights_in_losses(self):
        return self.cfg["training"]["use_class_weights_in_losses"]

    def get_optimizer_settings(self):
        return self.cfg["model"]["optimizer"]

    def get_loss_settings(self):
        return self.cfg["model"]["losses"]

    def get_metric_settings(self):
        return self.cfg["model"]["metrics"]

    def get_callback_settings(self):
        return self.cfg["model"]["callbacks"]

    def get_loss_weights(self):
        if "loss_weights" in self.cfg["model"]:
            loss_weights = self.cfg["model"]["loss_weights"]
        else:
            loss_weights = None
        return loss_weights

    def get_evaluation_metric_settings(self):
        return self.cfg["evaluation"]["metrics"]

    # ===============================================================
    # NEW
    # ===============================================================
    def get_preprocessings_settings(self):
        return self.cfg["preprocessing"]

    def get_evaluation_mask_labels(self):
        if "image_outputs" in self.cfg["evaluation"]:
            mask_labels = self.cfg["evaluation"]["image_outputs"]
            if mask_labels is None:
                mask_labels = []
        else:
            mask_labels = []

        return mask_labels

    def get_combine_patch_predictions(self):
        if "combine_patch_predictions" in self.cfg["evaluation"]:
            combine_patch_predictions = self.cfg["evaluation"]["combine_patch_predictions"]
        else:
            combine_patch_predictions = False

        return combine_patch_predictions

    def get_patch_predictions(self):
        if "patch_predictions" in self.cfg["evaluation"]:
            patch_predictions = self.cfg["evaluation"]["patch_predictions"]
        else:
            patch_predictions = True

        return patch_predictions

    def get_evaluate_train_set(self):
        if "evaluate_train_set" in self.cfg["evaluation"]:
            evaluate_train_set = self.cfg["evaluation"]["evaluate_train_set"]
        else:
            evaluate_train_set = True

        return evaluate_train_set

    def get_label_combination_type(self):
        if "combination_type" in self.cfg["evaluation"]:
            combination_type = self.cfg["evaluation"]["combination_type"]
        else:
            combination_type = None
        return combination_type

    def get_write_predictions(self):
        if "write_predictions" in self.cfg["evaluation"]:
            write_predictions = self.cfg["evaluation"]["write_predictions"]
        else:
            write_predictions = False
        return write_predictions

    def get_evaluate_metrics(self):
        if "evaluate_metrics" in self.cfg["evaluation"]:
            evaluate_metrics = self.cfg["evaluation"]["evaluate_metrics"]
        else:
            evaluate_metrics = True
        return evaluate_metrics

    def get_custom_definitions_file(self):
        file_found = False
        if "custom_definitions_file" in self.cfg["general"]:
            custom_definitions_file = self.cfg["general"]["custom_definitions_file"]
            if not os.path.exists(custom_definitions_file):
                custom_definitions_file = os.path.join(
                    self.cfg["original_location"], custom_definitions_file
                )

            if os.path.exists(custom_definitions_file):
                file_found = True

            if not file_found:
                raise ValueError("Custom definitions file was not found!")
        else:
            custom_definitions_file = None

        return custom_definitions_file

    def get_do_augmentation(self):
        if "data_augmentation" in self.cfg["training"]:
            return self.cfg["training"]["data_augmentation"]
        else:
            return False

    def get_data_augmentation_settings(self):
        if "augmentation_settings" in self.cfg["training"]:
            return self.cfg["training"]["augmentation_settings"]
        else:
            return {}

    def get_shuffle_val(self):
        if "shuffle_validation" in self.cfg["training"]:
            return self.cfg["training"]["shuffle_validation"]
        else:
            return False

    def get_class_weights(self):
        if "class_weights" in self.cfg["training"]:
            return self.cfg["training"]["class_weights"]
        else:
            return None

    def get_shuffle_evaluation(self):
        if "shuffle" in self.cfg["evaluation"]:
            shuffle_evaluation = self.cfg["evaluation"]["shuffle"]
        else:
            shuffle_evaluation = False

        return shuffle_evaluation

    def get_resume_training_from_model(self):
        if "resume_training" in self.cfg["training"]:
            return self.cfg["training"]["resume_training"]
        else:
            return False

    def get_model_file(self):
        if "model_file" in self.cfg["training"]:
            return self.cfg["training"]["model_file"]
        else:
            return None



import copy

from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Union

import numpy as np
import SimpleITK as sitk

from PrognosAIs.IO import utils as IO_utils


class ImageSample(ABC):
    """
    ImageSample base class

    To be implemented by subclasses:

        * `init_image_files`: Contains logic for retrieval of channel filepaths
        * `load_channels`: Contains logic for loading of channels from filepaths
        * `load_masks`: Contains logic of loading masks from filepaths

    Args:
        root_path: Path of the sample.
            Should contain folders or directories of channels and masks
        extension_keyword: Extension of the files to load
        mask_keyword (optional): Keyword to identify which filepaths are masks.
            Defaults to None.
    """

    def __init__(
        self,
        root_path: str,
        extension_keyword: str = None,
        mask_keyword: str = None,
        labels: dict = None,
        number_of_label_classes: dict = None,
        are_labels_one_hot: bool = False,
    ):
        self.root_path = copy.deepcopy(root_path)
        if isinstance(labels, dict):
            self.labels = [copy.deepcopy(labels)]
        elif isinstance(labels, list):
            self.labels = copy.deepcopy(labels)
        else:
            self.labels = [{}]

        if number_of_label_classes is not None:
            self.number_of_label_classes = copy.deepcopy(number_of_label_classes)
        else:
            self.number_of_label_classes = {}
        self.are_labels_one_hot = copy.deepcopy(are_labels_one_hot)
        self.mask_keyword = mask_keyword
        self.sample_name = IO_utils.get_root_name(self.root_path)

        if extension_keyword is not None:
            self.image_extension = extension_keyword
        self.image_files = self.init_image_files()

        self.channel_files = self._init_channel_files(self.image_files)
        self.mask_files = self._init_mask_files(self.image_files)
        self.number_of_channels = len(self.channel_files)
        self.number_of_masks = len(self.mask_files)

        self._channels = self.load_channels(self.channel_files)
        self._masks = self.load_masks(self.mask_files)

        self.has_patches = False
        self.number_of_patches = 0
        self.has_masks = self.number_of_masks > 0
        self._channel_patches = False
        self._mask_patches = False

        self.update_channel_size()
        self.update_mask_size()

        self._perform_sanity_checks()

        example_image = self.get_example_channel()
        self.original_metadata = {
            "image_size": example_image.GetSize(),
            "image_spacing": example_image.GetSpacing(),
            "image_origin": example_image.GetOrigin(),
            "image_direction": example_image.GetDirection(),
            "image": example_image,
        }
        self.update_metadata()
        self.update_labels()

    def copy(self):
        """
        Returns a (deep) copy of the instance

        Returns:
            ImageSample: Deep copy of the instance
        """

        return self.__class__(
            root_path=self.root_path,
            extension_keyword=self.image_extension,
            mask_keyword=self.mask_keyword,
            labels=self.labels,
            number_of_label_classes=self.number_of_label_classes,
            are_labels_one_hot=self.are_labels_one_hot,
        )

    def _perform_sanity_checks(self):
        """
        Automatic sanity check to see if we can process the sample

        Raises:
            NotImplementedError: If the configuration has not been implemented
        """

        if self.number_of_masks not in [0, 1, self.number_of_channels]:
            err_str = """Expect the number of masks to either be 0, 1 or equal to the number
             of channels but got {}!\nNumber of channels: {}"""
            raise NotImplementedError(err_str.format(self.number_of_masks, self.number_of_channels))

    def _identify_mask_file(self, image_file: str) -> bool:
        """
        Identify whether an image file is a mask based on the mask keyword

        Args:
            image_file: Image file to check

        Returns:
            bool: True if image_file is mask, False otherwise
        """

        if self.mask_keyword is None:
            is_mask_file = False
        else:
            is_mask_file = self.mask_keyword in IO_utils.get_root_name(image_file)

        return is_mask_file

    @abstractmethod
    def init_image_files(self) -> list:
        """
        Get the filepaths (folders or files) of the channels for a single sample.
        To be implemented by the subclass

        Returns:
            list: The filepaths of the channels
        """

    @abstractmethod
    def load_channels(self, channel_files: list) -> dict:
        """
        Load the channels from the channel files.
        To be implemented by the subclass

        Example subclass implementation::

            def load_channels(self, channel_files):
                channels = {}

                nifti_reader = sitk.ImageFileReader()
                nifti_reader.SetImageIO("NiftiImageIO")
                for i_channel_file in channel_files:
                    nifti_reader.SetFileName(i_channel_file)
                    i_channel = nifti_reader.Execute()
                    i_channel_name = os.path.basename(i_channel_file)
                    channels[i_channel_name] = i_channel
                return channels

        Args:
            channel_files (list): Paths to the channels to be loaded

        Returns:
            dict: mapping the channel file to the loaded image
        """

    @abstractmethod
    def load_masks(self, mask_files: list) -> dict:
        """
        Load the masks from the mask files.
        To be implemented by the subclass

        Example subclass implementation::

            def load_masks(self, mask_files):
                masks = {}
                nifti_reader = sitk.ImageFileReader()
                nifti_reader.SetImageIO("NiftiImageIO")
                for i_mask_file in mask_files:
                    nifti_reader.SetFileName(i_mask_file)
                    i_mask = nifti_reader.Execute()
                    i_mask = sitk.Cast(i_mask, sitk.sitkUInt8)
                    i_mask_name = IO_utils.get_file_name(i_mask_file, self.image_extension)
                    masks[i_mask_name] = i_mask
                return masks

        Args:
            mask_files (list): Paths to the masks to be loaded

        Returns:
            dict: mapping the mask file to the loaded mask
        """

    def _init_channel_files(self, image_files: list) -> list:
        """
        Get only the channel files from the image files, filtering out masks.

        Args:
            image_files (list): Paths to the image files

        Returns:
            list: The paths to the channel files
        """

        channel_files = [
            i_image_file
            for i_image_file in image_files
            if not self._identify_mask_file(i_image_file)
        ]
        return channel_files

    def _init_mask_files(self, image_files: list) -> list:
        """
        Get only the mask files from the image files, filtering out channels.

        Args:
            image_files (list): Paths to the image files

        Returns:
            list: The paths to the mask files
        """
        mask_files = [
            i_image_file for i_image_file in image_files if self._identify_mask_file(i_image_file)
        ]
        return mask_files

    @property
    def channel_size(self) -> np.ndarray:
        """
        The image size of the channels
        """
        return self._channel_size

    @channel_size.setter
    def channel_size(self, channel_size):
        self._channel_size = np.asarray(channel_size)

    def update_channel_size(self):
        """
        Update the channel size according to the current channels
        """
        example_channel = self.get_example_channel()
        if example_channel is not None:
            self.channel_size = example_channel.GetSize()
            self.number_of_dimensions = len(example_channel.GetSize())
        else:
            self.channel_size = None
            self.number_of_dimensions = None

    @property
    def mask_size(self) -> Union[np.ndarray, None]:
        """The image size of the masks."""
        if self.has_masks:
            return self._mask_size

        return None

    @mask_size.setter
    def mask_size(self, mask_size) -> None:
        if self.has_masks:
            self._mask_size = np.asarray(mask_size)

    def update_mask_size(self) -> None:
        """
        Update the mask size according to the current masks
        """
        if self.has_masks:
            example_mask = self.get_example_mask()
            if example_mask is not None:
                self.mask_size = example_mask.GetSize()
            else:
                self.mask_size = None

    def update_metadata(self) -> None:
        example_patches = self.get_example_channel_patches()
        self.metadata = []

        for i_patch in example_patches:
            # If the dimensionality changes then we need to do a little trick
            # we always just get the first dimension
            # TODO allow recontrsuction from multi-dimensional images
            if self.original_metadata["image"].GetDimension() > i_patch.GetDimension():
                patch_index = self.original_metadata["image"].TransformPhysicalPointToIndex(
                    i_patch.GetOrigin() + (0,),
                )
            else:
                patch_index = self.original_metadata["image"].TransformPhysicalPointToIndex(
                    i_patch.GetOrigin(),
                )

            i_metadata = {
                "patch_size": i_patch.GetSize(),
                "patch_direction": i_patch.GetDirection(),
                "patch_origin": i_patch.GetOrigin(),
                "patch_spacing": i_patch.GetSpacing(),
                "patch_index": patch_index,
            }

            self.metadata.append(i_metadata)

    def update_labels(self) -> None:
        if len(self.labels) < self.number_of_patches:
            new_labels = []
            for _ in range(self.number_of_patches):
                new_labels.append(copy.deepcopy(self.labels[0]))
            self.labels = new_labels

    def add_to_labels(
        self, to_add_labels: List[dict], to_add_number_of_label_classes: dict
    ) -> None:
        # If the user parses just a dict, we fix this here
        if isinstance(to_add_labels, dict):
            to_add_labels = [to_add_labels]

        if len(to_add_labels) != 1 and len(to_add_labels) != self.number_of_patches:
            err_msg = (
                "Labels to add should have either length 1 or length equal to"
                " number of patches.\n"
                "Number of patches is {}, but you have supplied {} labels.\n"
                "Please check your code and try again."
            ).format(self.number_of_patches, len(to_add_labels))
            raise ValueError(err_msg)

        if len(to_add_labels) == 1 and self.has_patches:
            to_add_labels = to_add_labels * self.number_of_patches

        for i_label, i_to_add_label in zip(self.labels, to_add_labels):
            i_label.update(i_to_add_label)

        self.number_of_label_classes.update(to_add_number_of_label_classes)

    def _parse_function_parameters(self, function_parameters):
        """
        Parse the function parameters

        Args:
            function_parameters (function or tuple): Function and possible args
                and kw_args.

        Returns:
            tuple: function, args, and kw_args
        """
        if isinstance(function_parameters, tuple):
            if len(function_parameters) == 2:
                function = function_parameters[0]
                if isinstance(function_parameters[1], list):
                    arguments = function_parameters[1]
                    kw_arguments = {}
                elif isinstance(function_parameters[1], dict):
                    arguments = []
                    kw_arguments = function_parameters[1]
            elif len(function_parameters) == 3:
                function = function_parameters[0]
                if isinstance(function_parameters[1], list):
                    arguments = function_parameters[1]
                    kw_arguments = function_parameters[2]
                else:
                    arguments = function_parameters[2]
                    kw_arguments = function_parameters[1]
        else:
            function = function_parameters
            arguments = []
            kw_arguments = {}

        return function, arguments, kw_arguments

    @property
    def channels(self) -> list:
        """
        The channels present in the sample

        Channels of a sample can be set by providing either a function,
         or a tuple consisting of a function, possible function argument
         and function keyword arguments.
        This function will then be applied to all channels in the sample.
        The function has to output either a SimpleITK Image or a list.
        In the last case it is assumed that these are patches and the class is updated accordingly


        Returns:
            list: Channels present in the sample
        """
        return [value for key, value in sorted(self._channels.items(), key=lambda item: item[0])]

    @channels.setter
    def channels(self, function_parameters):
        (function, function_arguments, function_kw_arguments,) = self._parse_function_parameters(
            function_parameters
        )

        example_function_output = function(
            self.get_example_channel(), *function_arguments, **function_kw_arguments
        )
        returns_patches = isinstance(example_function_output, list)

        if self._channel_patches and not returns_patches:
            for i_channel_name in self.channel_names:
                for i_i_patch, i_patch in enumerate(self._channels[i_channel_name]):
                    self._channels[i_channel_name][i_i_patch] = function(
                        i_patch, *function_arguments, **function_kw_arguments
                    )
        else:
            channel_items = sorted(self._channels.items(), key=lambda item: item[0])
            if not returns_patches:
                # We already have the first example, so we can use it for efficiency
                # If the function does return patches we should pass all patches at once
                # So we still need to calculate the output
                self._channels[channel_items[0][0]] = example_function_output
                channel_items = channel_items[1:]
            for i_i_channel, (i_channel_name, i_channel) in enumerate(channel_items):
                self._channels[i_channel_name] = function(
                    i_channel, *function_arguments, **function_kw_arguments
                )

        if returns_patches:
            self.has_patches = True
            self._channel_patches = True
            self.number_of_patches = len(self.get_example_channel_patches())
            self._number_of_channel_patches = len(self.get_example_channel_patches())

        self.update_channel_size()
        self.update_metadata()
        self.update_labels()

    @property
    def masks(self):
        """
        The masks present in the sample

        Masks of a sample can be set by providing either a function,
         or a tuple consisting of a function, possible function argument
         and function keyword arguments.
        This function will then be applied to all masks in the sample.
        The function has to output either a SimpleITK Image or a list.
        In the last case it is assumed that these are patches and the class is updated accordingly


        Returns:
            list: masks present in the sample
        """
        return [value for key, value in sorted(self._masks.items(), key=lambda item: item[0])]

    @masks.setter
    def masks(self, function_parameters):
        if self.has_masks:
            (
                function,
                function_arguments,
                function_kw_arguments,
            ) = self._parse_function_parameters(function_parameters)

            example_function_output = function(
                self.get_example_mask(), *function_arguments, **function_kw_arguments
            )
            returns_patches = isinstance(example_function_output, list)

            if self._mask_patches and not returns_patches:
                for i_mask_name in self.mask_names:
                    for i_i_patch, i_patch in enumerate(self._masks[i_mask_name]):
                        self._masks[i_mask_name][i_i_patch] = function(
                            i_patch, *function_arguments, **function_kw_arguments
                        )
            else:
                for i_mask_name, i_mask in self._masks.items():
                    self._masks[i_mask_name] = function(
                        i_mask, *function_arguments, **function_kw_arguments
                    )

            if returns_patches:
                self.has_patches = True
                self._mask_patches = True
                self._number_of_mask_patches = len(self.get_example_channel_patches())
                self.number_of_patches = len(self.get_example_mask_patches())
            self.update_mask_size()
            self.update_metadata()
            self.update_labels()

    @property
    def channel_names(self) -> list:
        """
        Names of the channels

        Returns:
            list: Channel names
        """
        return [key for key in sorted(self._channels.keys())]

    @property
    def mask_names(self) -> list:
        """
        Names of the masks

        Returns:
            list: Mask names
        """
        return [key for key in sorted(self._masks.keys())]

    def get_example_channel(self) -> sitk.Image:
        """
        Provides an example channel of the samples

        Returns:
            sitk.Image: Single channel of the sample
        """

        example = self.channels[0]
        if self._channel_patches and self._number_of_channel_patches > 0:
            example = example[0]
        elif self._channel_patches:
            # If we removed all patches, we will return None
            return None
        return sitk.Image(example)

    def get_example_mask(self) -> sitk.Image:
        """
        Provides an example mask of the samples

        Returns:
            sitk.Image: Single mask of the sample
        """

        if self.has_masks:
            example = self.masks[0]
            if self._mask_patches and self._number_of_mask_patches > 0:
                example = example[0]
            elif self._mask_patches:
                return None
            example = sitk.Image(example)
        else:
            example = None
        return example

    def get_example_channel_patches(self) -> list:
        """
        Provides an example of all patches of a channel, even if there is only one patch

        Returns:
            list: Patch(es) of a single channel of the sample
        """

        temp_example = self.channels[0]
        if not self._channel_patches:
            temp_example = [temp_example]

        # We ensure we make a copy so no operations are done on the
        # real sample
        example = []
        for i_patch in temp_example:
            example.append(sitk.Image(i_patch))
        return example

    def get_example_mask_patches(self) -> list:
        """
        Provides an example of all patches of a mask, even if there is only one patch

        Returns:
            list: Patch(es) of a single mask of the sample
        """

        if self.has_masks:
            temp_example = self.masks[0]
            if not self._mask_patches:
                temp_example = [temp_example]

            # We ensure we make a copy so no operations are done on the
            # real sample
            example = []
            for i_patch in temp_example:
                example.append(sitk.Image(i_patch))
        else:
            example = [None]
        return example

    def assert_all_channels_same_size(self):
        """
        Check wheter all channels have the same size

        Raises:
            ValueError: Raised when not all channels have same size
        """

        default_size = self.channel_size

        same_size = []
        for i_channel in self.channels:
            if self.has_patches:
                i_channel = i_channel[0]
            same_size.append(all(i_channel.GetSize() == default_size))

        if not all(same_size):
            raise ValueError("Not all channels have the same size, cannot continue!")

    def assert_all_masks_same_size(self):
        """
        Check wheter all masks have the same size

        Raises:
            ValueError: Raised when not all masks have same size
        """

        if self.has_masks:
            default_size = self.mask_size

            same_size = []
            for i_mask in self.masks:
                if self.has_patches:
                    i_mask = i_mask[0]
                same_size.append(all(i_mask.GetSize() == default_size))

            if not all(same_size):
                raise ValueError("Not all masks have the same size, cannot continue!")

    def get_grouped_channels(self) -> list:
        """
        Groups the channels on a per-patch basis instead of a per-channel basis

        The channels property indexes first by channel and then by (possibly) patches.
        This function instead first indexes by patches (or the whole sample of no patches).
        This can be handy when all channels are needed at the same time

        Returns:
            list: Grouped channels for each patch
        """

        grouped_channels = []
        if self.has_patches:
            channels = self.channels
            for i_patch in range(self.number_of_patches):
                grouped_channels.append(
                    [channels[i_channel][i_patch] for i_channel in range(self.number_of_channels)]
                )
        elif self.channels:
            grouped_channels.append(self.channels)
        else:
            # We dont have any channels
            grouped_channels = []

        return grouped_channels

    def get_grouped_masks(self) -> list:
        """
        Groups the masks on a per-patch basis instead of a per-channel basis

        The masks property indexes first by channel and then by (possibly) patches.
        This function instead first indexes by patches (or the whole sample of no patches).
        This can be handy when all masks are needed at the same time

        Returns:
            list: Grouped channels for each patch. Empty lists if sample doesn't have mask
        """
        grouped_masks = []
        if self.has_patches:
            masks = self.masks
            for i_patch in range(self.number_of_patches):
                grouped_masks.append(
                    [masks[i_mask][i_patch] for i_mask in range(self.number_of_masks)]
                )
        else:
            grouped_masks.append(self.masks)

        return grouped_masks

    @staticmethod
    def get_sitk_type_from_numpy_type(numpy_type: np.dtype) -> int:
        conversion_table = {
            "uint8": sitk.sitkUInt8,
            "uint16": sitk.sitkUInt16,
            "uint32": sitk.sitkUInt32,
            "uint64": sitk.sitkUInt64,
            "int8": sitk.sitkInt8,
            "int16": sitk.sitkInt16,
            "int32": sitk.sitkInt32,
            "int64": sitk.sitkInt64,
            # Simple ITK doesnt support float16
            "float16": sitk.sitkFloat32,
            "float32": sitk.sitkFloat32,
            "float64": sitk.sitkFloat64,
        }

        return conversion_table[str(numpy_type)]

    @staticmethod
    def get_numpy_type_from_sitk_type(sitk_type: int) -> np.dtype:
        conversion_table = {
            sitk.sitkUInt8: "uint8",
            sitk.sitkUInt16: "uint16",
            sitk.sitkUInt32: "uint32",
            sitk.sitkUInt64: "uint64",
            sitk.sitkInt8: "int8",
            sitk.sitkInt16: "int16",
            sitk.sitkInt32: "int32",
            sitk.sitkInt64: "int64",
            # Simple ITK doesnt support float16
            sitk.sitkFloat32: "float32",
            sitk.sitkFloat64: "float64",
        }

        return conversion_table[sitk_type]

    @staticmethod
    def get_appropiate_dtype_from_scalar(
        value: Union[int, float], return_np_type: bool = False
    ) -> Union[int, np.dtype]:
        """
        Find the minimum SimpleITK type need to represent the value

        Args:
            value (float): The value to check
            return_np_type (bool): If True returns the numpy type instead of the SimpleITK type.
             Defaults to False.

        Returns:
            int: The appropiate SimpleITK to which the value can be casted
        """

        # Add 1e10 to make sure that very large values are still converted to floats
        # Otherwise we have a problem when the value is larger than the maximum value
        # of a uint64
        if isinstance(value, float) and value.is_integer() and value < 1e10:
            value = int(value)

        appropiate_dtype = np.min_scalar_type(value)
        if not return_np_type:
            appropiate_dtype = ImageSample.get_sitk_type_from_numpy_type(appropiate_dtype)

        return appropiate_dtype

    @staticmethod
    def get_appropiate_dtype_from_image(image: sitk.Image) -> int:
        """
        Find the minimum SimpleITK type need to represent the value

        Args:
            value (float): The value to check

        Returns:
            int: The appropiate SimpleITK to which the value can be casted
        """

        datatypes = [
            "uint8",
            "uint16",
            "uint32",
            "uint64",
            "int8",
            "int16",
            "int32",
            "int64",
            "float16",
            "float32",
            "float64",
        ]

        image_array = sitk.GetArrayViewFromImage(image)
        min_image_value = np.amin(image_array)
        max_image_value = np.amax(image_array)

        has_only_integers = np.array_equal(image_array, image_array.astype(np.int))

        if has_only_integers:
            min_image_value = int(min_image_value)
            max_image_value = int(max_image_value)
            min_type = ImageSample.get_appropiate_dtype_from_scalar(
                min_image_value, return_np_type=True
            )
            # If the max value also fits within this type we are done.
            # Otherwise we might get too big of a type (e.g. when min = -10 and max = 10, it will give int16 instead of int8)
            # when promoting types
            if np.can_cast(max_image_value, min_type):
                data_type = min_type
            else:
                test_array = np.asarray([min_image_value, max_image_value])
                for i_data_type in datatypes:
                    if np.array_equal(test_array, test_array.astype(i_data_type)):
                        data_type = i_data_type
                        break
        else:
            # In the case of floats we need to make sure that the  whole range fits in a float
            if min_image_value < 0:
                value_range = max_image_value - min_image_value
            else:
                value_range = max_image_value

            range_data_type = ImageSample.get_appropiate_dtype_from_scalar(
                value_range, return_np_type=True
            )
            # If the min/max values are int, here we make sure that we cast it to at least a float
            data_type = np.promote_types(range_data_type, np.float32)

        return ImageSample.get_sitk_type_from_numpy_type(data_type)

    @staticmethod
    def promote_simpleitk_types(type_1: int, type_2: int) -> int:
        """
        Get the datatype that can represent both datatypes

        Args:
            type_1 (int): SimpleITK datype of variable 1
            type_2 (int): SimpleITK datatype of variable 2

        Returns:
            int: SimpleITK datatype that can represent both datatypes
        """

        np_type_1 = ImageSample.get_numpy_type_from_sitk_type(type_1)
        np_type_2 = ImageSample.get_numpy_type_from_sitk_type(type_2)

        np_common_type = np.promote_types(np_type_1, np_type_2)

        sitk_common_type = ImageSample.get_sitk_type_from_numpy_type(np_common_type)

        return sitk_common_type


class NIFTISample(ImageSample):
    def __init__(self, **kwds):
        self.image_extension = ".nii"
        super().__init__(**kwds)

    def load_channels(self, channel_files):
        channels = {}

        nifti_reader = sitk.ImageFileReader()
        nifti_reader.SetImageIO("NiftiImageIO")
        for i_channel_file in channel_files:
            nifti_reader.SetFileName(i_channel_file)
            i_channel = nifti_reader.Execute()
            i_channel_name = IO_utils.get_file_name(i_channel_file, self.image_extension)
            channel_dtype = self.get_appropiate_dtype_from_image(i_channel)
            # Channel has to be at minimum float32 because of pre-processing operations
            channel_dtype = self.promote_simpleitk_types(channel_dtype, sitk.sitkFloat32)
            i_channel = sitk.Cast(i_channel, channel_dtype)
            channels[i_channel_name] = i_channel
        return channels

    def load_masks(self, mask_files):
        masks = {}
        nifti_reader = sitk.ImageFileReader()
        nifti_reader.SetImageIO("NiftiImageIO")
        for i_mask_file in mask_files:
            nifti_reader.SetFileName(i_mask_file)
            i_mask = nifti_reader.Execute()
            mask_dtype = self.get_appropiate_dtype_from_image(i_mask)
            i_mask = sitk.Cast(i_mask, mask_dtype)
            i_mask_name = IO_utils.get_file_name(i_mask_file, self.image_extension)
            masks[i_mask_name] = i_mask
        return masks

    def init_image_files(self):
        image_files = IO_utils.find_files_with_extension(self.root_path, self.image_extension)
        return image_files


def get_sample_class(sample_type_name: str):
    if sample_type_name == "nifti":
        return NIFTISample
